#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 13 2023
@author: Bridget Hass (bhass@battelleecology.org)

Functions to download AOP data either for an entire site (by_file_aop), or
for a portion of a site (by_tile_aop), specified by the UTM coordinates
(easting and northing) of the AOP data tiles that you want to download.

Adapted from R neonUtilities byFileAOP.R and byTileAOP.R, along with other helper functions.
https://github.com/NEONScience/NEON-utilities/blob/main/neonUtilities/R/byFileAOP.R
https://github.com/NEONScience/NEON-utilities/blob/main/neonUtilities/R/byTileAOP.R
written by:
@author: Claire Lunch (clunch@battelleecology.org)
@author: Christine Laney (claney@battelleecology.org)

Updated Oct 2025 to include options to skip files if they already exist locally and 
to validate existing files against files published on the Data Portal with checksums.

"""

# %% imports
from datetime import datetime
import glob
import importlib_resources
import logging
import pandas as pd
import numpy as np
import os
import re

# import time
from time import sleep
from tqdm import tqdm

# local imports
from . import __resources__
from .helper_mods.api_helpers import get_api
from .helper_mods.api_helpers import token_check
from .helper_mods.api_helpers import download_file, calculate_crc32c
from .helper_mods.metadata_helpers import convert_byte_size
from .get_issue_log import get_issue_log
from .citation import get_citation

# display the log info messages, only showing the message (otherwise it would print INFO:root:'message')
logging.basicConfig(level=logging.INFO, format="%(message)s")


# check that token was used
def check_token(response):
    """
    This function checks the API response headers for a rate limit. If the rate limit is found to be '200',
    it prints a warning message indicating that the API token was not recognized and the public rate limit was applied.

    Parameters
    --------
    response: API get response object containing status code and json data. Return of get_api function.

    Returns
    --------
    None

    Raises
    --------
    None

    """

    if (
        "x-ratelimit-limit" in response.headers
        and response.headers["x-ratelimit-limit"] == "200"
    ):
        logging.info("API token was not recognized. Public rate limit applied.\n")


# %%


def get_file_urls(urls, token=None):
    """
    This function retrieves all the files from a list of NEON data URLs.

    Parameters
    --------
    urls: list of str
        A list of URLs from which the files will be retrieved.

    token: str, optional
        User-specific API token generated within neon.datascience user accounts. If provided, it will be used for authentication.

    Returns
    --------
    all_file_url_df: pandas.DataFrame
        A DataFrame containing information about all the files retrieved from the URLs.
        The DataFrame includes columns for 'name', 'size', 'crc32c', and 'url' and 'release' of the files.

    release: str
        The release information retrieved from the response JSON.

    Raises
    --------
    None

    Examples
    --------
    >>> urls = ['https://data.neonscience.org/api/v0/data/DP3.30015.001/SCBI/2022-05',
                'https://data.neonscience.org/api/v0/data/DP3.30015.001/SCBI/2023-06']
    >>> get_file_urls(urls, 'my-api-token')
    # This will return a DataFrame with all the data files contained in SCBI 2022-05 and 2023-06 and the release information.

    Notes
    --------
    The function makes API calls to each URL in the 'urls' list and retrieves the file
    information. It also retrieves the release information from the response JSON.
    If the API call fails, it prints a warning message and continues with the next URL.
    """

    all_file_url_df = pd.DataFrame()
    releases = []
    for url in urls:
        response = get_api(api_url=url, token=token)
        if response is None:
            logging.info(
                "NEON data file retrieval failed. Check NEON data portal for outage alerts."
            )

        # get release info
        release = response.json()["data"]["release"]
        releases.append(release)

        file_url_dict = response.json()["data"]["files"]
        file_url_df = pd.DataFrame(data=file_url_dict)
        file_url_df["release"] = release

        # drop md5 and crc32 columns, which are all NaNs
        file_url_df.drop(columns=["md5", "crc32"], inplace=True)

        # append the new dataframe to the existing one
        all_file_url_df = pd.concat([all_file_url_df, file_url_df], ignore_index=True)

    return all_file_url_df, list(set(releases))


# %%


def get_shared_flights(site):
    """
    This function retrieves shared flights information for a NEON site from
    the shared_flights.csv lookup file, which contains the mapping between
    collocated AOP sites (terrestrial:terrestrial and aquatic:terrestrial).
    If the site is found in the shared flights data, it prints a message and
    updates the site to the corresponding "flightSite".

    Parameters
    --------
    site: str
        The 4-letter NEON site code for which the collocated site information is to be retrieved.

    Returns
    --------
    site: str
        The collocated site that AOP data is published under. If the original
        site is found in the shared flights lookup, it is updated to the
        corresponding flight site. Otherwise, it remains the same.

    Raises
    --------
    None

    Examples
    --------
    >>> get_shared_flights('TREE')
    'TREE is part of the flight box for STEI. Downloading data from STEI.'

    Notes
    --------
    The function reads the shared flights data from a CSV file named
    'shared_flights.csv' located in the '__resources__' directory.

    """
    shared_flights_file = (
        importlib_resources.files(__resources__) / "shared_flights.csv"
    )

    shared_flights_df = pd.read_csv(shared_flights_file)

    shared_flights_dict = shared_flights_df.set_index(["site"])["flightSite"].to_dict()
    if site in shared_flights_dict:
        flightSite = shared_flights_dict[site]
        if site in ["TREE", "CHEQ", "KONA", "DCFS"]:
            logging.info(
                f"{site} is part of the NEON flight box for {flightSite}. Downloading data from {flightSite}."
            )
        else:
            logging.info(
                f"{site} is a NEON aquatic site and is sometimes included in the flight box for {flightSite}. Aquatic sites are not always included in the flight coverage every year.\nDownloading data from {flightSite}. Check data to confirm coverage of {site}."
            )
        site = flightSite
    return site


def get_neon_sites():
    """This function gets a list of the valid NEON sites from the
    neon_field_site_metadata.csv file for validation, and adds the AOP CHEQ
    site, which is an AOP site name and is associated with STEI & TREE."
    """
    neon_sites_file = (
        importlib_resources.files(__resources__)
        / "neon_field_site_metadata_20250214.csv"
    )
    neon_sites_df = pd.read_csv(neon_sites_file)

    neon_sites_list = list(neon_sites_df["field_site_id"])
    neon_sites_list.append(
        "CHEQ"
    )  # append the CHEQ site, a separate flight area than STEI-TREE

    return neon_sites_list


def get_data_product_name(dpid):
    dpid_api_response = get_api(f"https://data.neonscience.org/api/v0/products/{dpid}")
    product_name = dpid_api_response.json()["data"]["productName"]
    return product_name


def check_exists_and_checksum(row, download_path):
    """
    Check if a file exists locally and if its CRC32C checksum matches the expected value.
    Used in the skip_if_exists option of by_tile_aop and by_file_aop.

    Parameters
    ----------
    row : pandas.Series
        A row from a DataFrame containing at least a 'url' and optionally a 'crc32c' field.
    download_path : str
        The root directory where files are downloaded.

    Returns
    -------
    pandas.Series
        A Series of two boolean values:
        - exists_locally: True if the file exists locally, False otherwise.
        - checksum_matches: True if the local file's CRC32C matches the expected value, False otherwise.

    Example
    -------
    >>> file_url_df[["exists_locally", "checksum_matches"]] = file_url_df.apply(
    ...     lambda row: check_exists_and_checksum(row, download_path), axis=1
    ... )
    """
    # print(row["url"])
    pathparts = row["url"].split("/")
    file_path = os.path.join(download_path, *pathparts[3:])
    # file_path = os.path.join(download_path,"\\".join(pathparts[3:]))
    # print(file_path)
    exists = os.path.exists(file_path)
    # print(exists)
    if not exists:
        # print(f'{pathparts[-1:]} does not exist!')
        return pd.Series([False, False])
    expected_crc32c = row.get("crc32c")
    # print(expected_crc32c)
    if expected_crc32c:
        local_crc32c = calculate_crc32c(file_path)
        matches = local_crc32c.zfill(8) == str(expected_crc32c).zfill(8)
        # return pd.Series({"exists_locally": True, "checksum_matches": matches})
        return pd.Series([True, matches])
    return pd.Series([True, False])
    # return pd.Series({"exists_locally": True, "checksum_matches": False})


# %% functions to validate inputs for by_file_aop and by_tile_aop


def validate_dpid(dpid):
    dpid_pattern = "DP[1-4]{1}.[0-9]{5}.00[1-2]{1}"
    if not re.fullmatch(dpid_pattern, dpid):
        raise ValueError(
            f"{dpid} is not a properly formatted NEON data product ID. The correct format is DP#.#####.00#"
        )


# %% function to get all AOP data product IDs (active = valid, inactive = future / suspended)


def get_aop_dpids():
    """
    This function retrieves all active and inactive AOP data product IDs from the NEON API.

    Returns
    -------
    active_dpids: list
        A list of all active AOP data product IDs (productStatus = "ACTIVE")

    inactive_dpids: list
        A list of all inactive AOP data product IDs (productStatus = "FUTURE" / suspended products)

    Raises
    ------
    None

    Examples
    --------
    >>> active_dpids = get_active_dpids()
    # This will return a list of all active NEON data product IDs.
    """
    response = get_api("https://data.neonscience.org/api/v0/products")

    response_dict = response.json()
    # all_neon_dpids = [item["productCode"] for item in response_dict["data"]]
    # active_neon_dpids = [item["productCode"] for item in response_dict["data"] if item["productStatus"] == "ACTIVE"]
    active_aop_dpids = [
        item["productCode"]
        for item in response_dict["data"]
        if (
            item["productScienceTeamAbbr"] == "AOP"
            and item["productStatus"] == "ACTIVE"
        )
    ]
    inactive_aop_dpids = [
        item["productCode"]
        for item in response_dict["data"]
        if (
            item["productScienceTeamAbbr"] == "AOP"
            and item["productStatus"] == "FUTURE"
        )
    ]

    return active_aop_dpids, inactive_aop_dpids


def validate_aop_dpid(dpid):
    """
    Validates the given AOP data product ID against a pattern and a list of active DPIDs, determined from the API.

    Parameters:
    - dpid (str): The data product ID to validate.

    Raises:
    - ValueError: If the dpid does not match the expected pattern or is not in the list of active (valid) DPIDs.
    """
    # Regular expression pattern for AOP DPIDs
    aop_dpid_pattern = "DP[1-3]{1}.300[0-2]{1}[0-9]{1}.00[1-2]{1}"

    # Check if the dpid matches the pattern
    if not re.fullmatch(aop_dpid_pattern, dpid):
        raise ValueError(
            f"{dpid} is not a valid NEON AOP data product ID. AOP data products follow the format DP#.300##.00#."
        )

    active_aop_dpids, inactive_aop_dpids = get_aop_dpids()

    # Check if the dpid is in the list of suspended (FUTURE) AOP DPIDs
    if dpid in inactive_aop_dpids:
        raise ValueError(
            f"NEON {dpid} has been suspended and is not currently available, see https://data.neonscience.org/data-products/{dpid} for more details."
        )

    # Check if the dpid is in the list of valid AOP DPIDs
    if dpid not in active_aop_dpids:
        active_aop_dpids.sort()
        valid_aop_dpids_string = "\n".join(active_aop_dpids)
        raise ValueError(
            f"NEON {dpid} is not a valid AOP data product ID. Valid AOP IDs are listed below:\n{valid_aop_dpids_string}"
        )


def validate_aop_l3_dpid(dpid):
    """
    Validates the given AOP data product ID against expected pattern to check if it is downloadable by tile.
    If the dpid does not match the expected pattern or is not in the list of active Level 3 AOP data product IDs,
    it will  raise a value error with a descriptive message.

    Parameters:
    - dpid (str): The data product ID to validate.

    Raises:
    - ValueError: If the dpid is not in the list of AOP data product IDs that are downloadable by tile.
    """
    # Check if the dpid starts with DP3 or is DP1.30003.001
    if not (dpid.startswith("DP3") or dpid == "DP1.30003.001"):
        raise ValueError(
            f"NEON {dpid} is not a valid Level 3 (L3) AOP data product ID. Level 3 AOP products follow the format DP3.300##.00#, with the exception of DP1.30003.001 (the discrete classified lidar point cloud data product)."
        )

    # Get the list of active AOP data product IDs
    active_aop_dpids, _ = get_aop_dpids()

    # List of valid Level 3 AOP data product IDs
    valid_aop_l3_dpids = [
        dpid
        for dpid in active_aop_dpids
        if dpid.startswith("DP3") or dpid == "DP1.30003.001"
    ]

    # Check if the dpid is in the list of valid AOP dpids
    if dpid not in valid_aop_l3_dpids:
        valid_aop_l3_dpids.sort()
        valid_aop_l3_dpids_string = "\n".join(valid_aop_l3_dpids)

        raise ValueError(
            f"NEON {dpid} is not a valid Level 3 (L3) AOP data product ID. Valid L3 AOP IDs are listed below:\n{valid_aop_l3_dpids_string}"
        )
        # below prints out the corresponding data product names for each ID.
        # f'{dpid} is not a valid Level 3 (L3) AOP data product ID. Valid L3 AOP products are listed below.\n{formatted_dpid_dict}')


def check_field_spectra_dpid(dpid):
    if dpid == "DP1.30012.001":
        raise ValueError(
            f"NEON {dpid} is the Field spectral data product, which is published as tabular data. Use zips_by_product() or loadByProduct() to download these data."
        )


def validate_site_format(site):
    site_pattern = "[A-Z]{4}"
    if not re.fullmatch(site_pattern, site):
        raise ValueError(
            f"{site} is an invalid NEON site format. A four-letter NEON site code is required. NEON site codes can be found here: https://www.neonscience.org/field-sites/explore-field-sites"
        )


def validate_neon_site(site):
    neon_sites = get_neon_sites()

    if site not in neon_sites:
        raise ValueError(
            f"{site} is not a valid NEON site code. A complete list of NEON site codes can be found here: https://www.neonscience.org/field-sites/explore-field-sites"
        )


def validate_year(year):
    # year = str(year)
    # year_pattern = "20?(1[2-9]|2[0-9])"
    # if not re.fullmatch(year_pattern, year):
    #     raise ValueError(
    #         f'{year} is an invalid year. Year is required in the format "2017" or 2017, eg. NEON AOP data are available from 2013 to present.'
    #     )
    """
    Validates that the year is between 2012 and the current year.
    """
    # First, check with regex for exactly 4 digits
    if not re.fullmatch(r"\d{4}", str(year)):
        raise ValueError(
            f"{year} is an invalid year. Year is required in the format '2017' or 2017, e.g. NEON AOP data are available from 2012 to present."
        )

    # Convert year to an integer and check if it's between 2012 and the current year
    year_int = int(year)
    current_year = datetime.now().year
    if not (2012 <= year_int <= current_year):
        raise ValueError(
            f"{year} is an invalid year. Year must be between 2012 and {current_year}."
        )


def validate_overwrite(overwrite):
    """
    Validates that overwrite is one of the accepted options: 'yes', 'no', or 'prompt'.
    Raises a ValueError if not.
    """
    valid_options = {"yes", "no", "prompt"}
    if overwrite not in valid_options:
        raise ValueError(f"overwrite must be one of {valid_options}. ")


def validate_skip_if_exists(skip_if_exists):
    """
    Validates that skip_if_exists is a boolean (True or False).
    Raises a ValueError if not.
    """
    if not isinstance(skip_if_exists, bool):
        raise ValueError(
            f"skip_if_exists must be a boolean (True or False). "
            f"Received: '{skip_if_exists}' of type {type(skip_if_exists)}"
        )


def check_aop_dpid(response_dict, dpid):
    if response_dict["data"]["productScienceTeamAbbr"] != "AOP":
        logging.info(
            f"NEON {dpid} is not a remote sensing product. Use zipsByProduct()"
        )
        return


def get_site_year_urls(response_dict, site, year):
    site_info = next(
        item for item in response_dict["data"]["siteCodes"] if item["siteCode"] == site
    )
    site_urls = site_info["availableDataUrls"]
    site_year_urls = [url for url in site_urls if str(year) in url]
    return site_year_urls


# %% functions to display available dates and tile extents


def list_available_dates(dpid, site):
    """
        list_available_dates displays the available releases and dates for a given product and site
        --------
         Inputs:
             dpid: the data product code (eg. 'DP3.30015.001' - CHM)
             site: the 4-digit NEON site code (eg. 'JORN')
        --------
        Returns:
        prints the Release Tag (or PROVISIONAL) and the corresponding available dates (YYYY-MM) for each tag
    --------
        Usage:
        --------
        >>> list_available_dates('DP3.30015.001','JORN')
        RELEASE-2025 Available Dates: 2017-08, 2018-08, 2019-08, 2021-08, 2022-09

        >>> list_available_dates('DP3.30015.001','HOPB')
        PROVISIONAL Available Dates: 2024-09
        RELEASE-2025 Available Dates: 2016-08, 2017-08, 2019-08, 2022-08

        >>> list_available_dates('DP1.10098.001','HOPB')
        ValueError: There are no data available for the data product DP1.10098.001 at the site HOPB.
    """
    product_url = "https://data.neonscience.org/api/v0/products/" + dpid
    response = get_api(api_url=product_url)  # add input for token?

    # raise value error and print message if dpid isn't formatted as expected
    validate_dpid(dpid)

    # raise value error and print message if site is not a 4-letter character
    site = site.upper()  # make site upper case (if it's not already)
    validate_site_format(site)

    # raise value error and print message if site is not a valid NEON site
    validate_neon_site(site)

    # check if product is active
    if response.json()["data"]["productStatus"] != "ACTIVE":
        raise ValueError(
            f"NEON {dpid} is not an active data product. See https://data.neonscience.org/data-products/{dpid} for more details."
        )

    # get available releases & months:
    for i in range(len(response.json()["data"]["siteCodes"])):
        if site in response.json()["data"]["siteCodes"][i]["siteCode"]:
            available_releases = response.json()["data"]["siteCodes"][i][
                "availableReleases"
            ]

    # display available release tags (including provisional) and dates for each tag
    try:
        for entry in available_releases:
            release = entry["release"]
            available_months = ", ".join(entry["availableMonths"])
            logging.info(f"{release} Available Dates: {available_months}")
    except UnboundLocalError:
        # if the available_releases variable doesn't exist, this error will show up:
        # UnboundLocalError: local variable 'available_releases' referenced before assignment
        raise ValueError(
            f"There are no NEON data available for the data product {dpid} at the site {site}."
        )


def get_tile_bounds(file_url_df, all_bounds=False):
    """
    Extracts and calculates the bounding coordinates from a DataFrame of file
    names containing UTM coordinates.The input dataframe can be generated from
    the function get_file_urls.

    This function filters out file names that end with the data extensions
    ('.tif', '.h5', '.laz', '.zip') and extracts UTM x and y coordinates from
    the data file names. It calculates the minimum and maximum x and y
    coordinates to determine the bounding box of the tiles. Additionally, it
    returns a sorted list of unique x_y coordinate pairs.

    Parameters:
    - file_url_df (pd.DataFrame): A DataFrame containing a 'name' column with
      file names that include UTM coordinates.
    - all_bounds (bool, optional): A flag to indicate whether to return a list
      of all UTM coordinates. Default is set to False.

    Returns:
    - x_bounds (tuple): A tuple containing the minimum and maximum x coordinates (min_x, max_x).
    - y_bounds (tuple): A tuple containing the minimum and maximum y coordinates (min_y, max_y).
    - sorted_coords (list): A sorted list of unique (x, y) coordinate tuples extracted from the file names.

    Example:
    >>> file_url_df, releases = get_file_urls(site_year_urls, token=token)
    >>> x_bounds, y_bounds, sorted_coords = get_tile_bounds(file_url_df)
    """

    # Regular expression to match UTM coordinates in the format 'xxxxxx_yyyyyyy'
    utm_pattern = re.compile(r"(\d{6})_(\d{7})")

    # lists to store x and y coordinates
    x_coords = []
    y_coords = []
    unique_coords = set()

    # filter out rows where 'name' ends with '.tif' , '.h5' or '.laz'
    # this will exclude shapefiles, just in case they don't match
    data_df = file_url_df[
        file_url_df["name"].str.endswith((".tif", ".h5", ".laz", ".zip"))
    ]

    # Iterate over each name in the DataFrame
    for name in data_df["name"]:
        match = utm_pattern.search(name)
        if match:
            x, y = match.groups()
            x_coords.append(int(x))
            y_coords.append(int(y))
            unique_coords.add((int(x), int(y)))

    # Convert the set to a sorted list
    sorted_coords = sorted(unique_coords)

    # Calculate min and max for x and y coordinates
    min_x = min(x_coords)
    max_x = max(x_coords)
    min_y = min(y_coords)
    max_y = max(y_coords)

    x_bounds = (min_x, max_x)
    y_bounds = (min_y, max_y)

    return x_bounds, y_bounds, sorted_coords


def get_aop_tile_extents(dpid, site, year, token=None):
    """
    This function displays the tile extents for a given product, site, and year
    and returns a complete list of the UTM coordinates

    Some NEON sites have irregular polygon flight plan boundaries, or may have
    incomplete coverage in a given year, so the list of coordinates provides a
    complete picture of the available tiles.

    Parameters
    --------
    dpid: str
        The identifier of the NEON data product to pull, in the form DPL.PRNUM.REV, e.g. DP3.30015.001.
        This must be a DP3 (Level 3) data product to work.

    site: str
        The four-letter code of a single NEON site, e.g. 'CLBJ'.

    year: str or int
        The four-digit year of data collection.

    token: str, optional
        User-specific API token from data.neonscience.org user account. See
        https://data.neonscience.org/data-api/rate-limiting/ for details about
        API rate limits and user tokens.

    Returns
    --------
    sorted_coords: list, a sorted list of the UTM x, y pairs of all tiles
    prints the minimum and maximum UTM x and y coordinates for the SW corners of the tiles encompassing the full site

    Examples
    --------
    >>> tile_extents = get_aop_tile_extents(dpid="DP3.30015.001", site="MCRA", year="2021",all_bounds=True)
    # This returns a list of the UTM x,y extent for all CHM tiles at the site MCRA collected in 2021.
    # It also displays the minimum and maximum UTM Easting and Northing (x and y) values for this product - site -year.

    """

    # raise value error and print message if dpid isn't formatted as expected
    validate_dpid(dpid)

    # raise value error and print message if dpid isn't a valid AOP L3 product
    validate_aop_l3_dpid(dpid)

    # raise value error and print message if field spectra data product is entered
    check_field_spectra_dpid(dpid)

    # raise value error and print message if site is not a 4-letter character
    site = site.upper()  # make site upper case (if it's not already)
    validate_site_format(site)

    # raise value error and print message if site is not a valid NEON site
    validate_neon_site(site)

    # raise value error and print message if year input is not valid
    year = str(year)  # cast year to string (if it's not already)
    validate_year(year)

    # if token is an empty string, set to None
    if token == "":
        token = None

    # query the products endpoint for the product requested
    response = get_api("https://data.neonscience.org/api/v0/products/" + dpid, token)

    # exit function if response is None (eg. if no internet connection)
    if response is None:
        logging.info("No response from NEON API. Check internet connection")
        return

    # check that token was used
    if token and "x-ratelimit-limit" in response.headers:
        check_token(response)
        # if response.headers['x-ratelimit-limit'] == '200':
        #     print('API token was not recognized. Public rate limit applied.\n')

    # get the request response dictionary
    response_dict = response.json()

    # error message if dpid is not an AOP data product
    check_aop_dpid(response_dict, dpid)

    # replace collocated site with the AOP site name it's published under
    site = get_shared_flights(site)

    # get the urls for months with data available, and subset to site & year
    site_year_urls = get_site_year_urls(response_dict, site, year)

    # error message if nothing is available
    if len(site_year_urls) == 0:
        logging.info(
            f"There are no NEON {dpid} data available at the site {site} in {year}. \nTo display available dates for a given data product and site, use the function list_available_dates()."
        )
        return

    # get file url dataframe for the available month urls
    file_url_df, releases = get_file_urls(site_year_urls, token=token)

    # get the number of files in the dataframe, if there are no files to download, return
    if len(file_url_df) == 0:
        # print("No data files found.")
        logging.info("No data files found.")
        return

    # corner_tiles = get_corner_tiles(file_url_df)
    x_bounds, y_bounds, sorted_coords = get_tile_bounds(file_url_df)

    logging.info(f"Easting Bounds: {x_bounds}")
    logging.info(f"Northing Bounds: {y_bounds}")

    # return the sorted_coords list
    return sorted_coords


# %%


def by_file_aop(
    dpid,
    site,
    year,
    include_provisional=False,
    check_size=True,
    savepath=None,
    chunk_size=1024,
    token=None,
    verbose=False,
    skip_if_exists=False,
    overwrite="prompt",
):
    """
    This function queries the NEON API for AOP data by site, year, and product, and downloads all
    files found, preserving the original folder structure. It downloads files serially to
    avoid API rate-limit overload, which may take a long time.

    Parameters
    --------
    dpid: str
        The identifier of the NEON data product to pull, in the form DPL.PRNUM.REV, e.g. DP3.30001.001.

    site: str
        The four-letter code of a single NEON site, e.g. 'CLBJ'.

    year: str or int
        The four-digit year of data collection.

    include_provisional: bool, optional
        Should provisional data be downloaded? Defaults to False. See
        https://www.neonscience.org/data-samples/data-management/data-revisions-releases
        for details on the difference between provisional and released data.

    check_size: bool, optional
        Should the user approve the total file size before downloading? Defaults to True.
        If you have sufficient storage space on your local drive, when working
        in batch mode, or other non-interactive workflow, use check_size=False.

    savepath: str, optional
        The file path to download to. Defaults to None, in which case the working directory is used.

    chunk_size: integer, optional
        Size in bytes of chunk for chunked download. Defaults to 1024.

    token: str, optional
        User-specific API token from data.neonscience.org user account. Defaults to None.
        See https://data.neonscience.org/data-api/rate-limiting/ for details about API
        rate limits and user tokens.

    verbose: bool, optional
        If set to True, the function will print more detailed information about the download process.

    skip_if_exists: bool, optional
        If set to True, the function will skip downloading files that already exist in the
        savepath and are valid (local checksums match the checksums of the published file).
        Defaults to False. If any local file checksums don't match those of files published
        on the NEON Data Portal, the user will be prompted to skip these files or overwrite
        the existing files with the new ones (see overwrite input).

    overwrite: str, optional
        Must be one of:
            'yes'    - overwrite mismatched files without prompting,
            'no'     - don't overwrite mismatched files (skip them, no prompt),
            'prompt' - prompt the user (y/n) to overwrite mismatched files after displaying them (default).
        If skip_if_exists is False, this parameter is ignored, and any existing files in
        the savepath will be overwritten according to the function's default behavior.

    Returns
    --------
    None; data are downloaded to the directory specified (savepath) or the current working directory.
    If data already exist in the expected path, they will be overwritten by default. To check for
    existing files before downloading, set skip_if_exists=True along with an overwrite option (y/n/prompt).

    Examples
    --------
    >>> by_file_aop(dpid="DP3.30015.001",
                    site="MCRA",
                    year="2021",
                    savepath="./test_download",
                    skip_if_exists=True)
    # This downloads the 2021 Canopy Height Model data from McRae Creek to the './test_download' directory.
    # If any files already exist in the savepath, they will be checked and skipped if they are valid.
    # The user will be prompted to ovewrite or skip downloading any existing files that do not match
    # the latest published data on the NEON Data Portal.

    Notes
    --------
    This function creates a folder named by the Data Product ID (DPID; e.g. DP3.30015.001) in the
    'savepath' directory, containing all AOP files meeting the query criteria. If 'savepath' is
    not provided, data are downloaded to the working directory, in a folder named by the DPID.
    """

    # raise value error and print message if dpid isn't formatted as expected
    validate_dpid(dpid)

    # raise value error and print message if dpid isn't formatted as expected
    validate_aop_dpid(dpid)

    # raise value error and print message if field spectra data are attempted
    check_field_spectra_dpid(dpid)

    # raise value error and print message if site is not a 4-letter character
    site = site.upper()  # make site upper case (if it's not already)
    validate_site_format(site)

    # raise value error and print message if site is not a valid NEON site
    validate_neon_site(site)

    # raise value error and print message if year input is not valid
    year = str(year)  # cast year to string (if it's not already)
    validate_year(year)

    # raise value error and print message if skip_if_exists input is not valid (boolean)
    validate_skip_if_exists(skip_if_exists)

    # raise value error and print message if validate_overwrite input is not valid (yes, no, prompt)
    validate_overwrite(overwrite)

    # warn if overwrite is set but skip_if_exists is False
    if not skip_if_exists and overwrite != "prompt":
        logging.info(
            "WARNING: overwrite option only applies if skip_if_exists=True. "
            "By default, any existing files will be overwritten unless you select skip_if_exists=True and overwrite='no' or 'prompt' (default)."
        )

    # if token is an empty string, set to None
    if token == "":
        token = None

    # check for expired token
    if token is not None:
        token = token_check(token)

    # query the products endpoint for the product requested
    response = get_api("https://data.neonscience.org/api/v0/products/" + dpid, token)

    # exit function if response is None (eg. if no internet connection)
    if response is None:
        logging.info("No response from NEON API. Check internet connection")
        return

    # check that token was used
    if token and "x-ratelimit-limit" in response.headers:
        check_token(response)

    # get the request response dictionary
    response_dict = response.json()

    # error message if dpid is not an AOP data product
    check_aop_dpid(response_dict, dpid)

    # replace collocated site with the AOP site name it's published under
    site = get_shared_flights(site)

    # get the urls for months with data available, and subset to site & year
    site_year_urls = get_site_year_urls(response_dict, site, year)

    # error message if nothing is available
    if len(site_year_urls) == 0:
        logging.info(
            f"There are no NEON {dpid} data available at the site {site} in {year}.\nTo display available dates for a given data product and site, use the function list_available_dates()."
        )
        # print("There are no data available at the selected site and year.")
        return

    # get file url dataframe for the available month urls
    file_url_df, releases = get_file_urls(site_year_urls, token=token)

    # get the number of files in the dataframe, if there are no files to download, return
    if len(file_url_df) == 0:
        logging.info("No NEON data files found.")
        return

    # if 'PROVISIONAL' in releases and not include_provisional:
    if include_provisional:
        # log provisional included message
        logging.info(
            "Provisional NEON data are included. To exclude provisional data, use input parameter include_provisional=False."
        )
    else:
        # log provisional not included message and filter to the released data
        file_url_df = file_url_df[file_url_df["release"] != "PROVISIONAL"]
        if len(file_url_df) == 0:
            logging.info(
                "Provisional NEON data are not included. To download provisional data, use input parameter include_provisional=True."
            )

    num_files = len(file_url_df)
    if num_files == 0:
        logging.info(
            "No NEON data files found. Available data may all be provisional. To download provisional data, use input parameter include_provisional=True."
        )
        return

    # get the total size of all the files found
    download_size_bytes = file_url_df["size"].sum()
    # print(f'download size, bytes: {download_size_bytes}')
    download_size = convert_byte_size(download_size_bytes)
    # print(f'download size: {download_size}')

    # report data download size and ask user if they want to proceed
    if check_size:
        if (
            input(
                f"Continuing will download {num_files} NEON data files totaling approximately {download_size}. Do you want to proceed? (y/n) "
            )
            .strip()
            .lower()  # lower or upper case 'y' will work
            != "y"
        ):
            print("Download halted.")
            return

    # create folder in working directory to put files in
    if savepath is not None:
        download_path = savepath + "/" + dpid
    else:
        download_path = os.getcwd() + "/" + dpid
    os.makedirs(download_path, exist_ok=True)

    # serially download all files, with progress bar
    files = list(file_url_df["url"])

    # different messages depending on whether skip_if_exists is True or False
    if skip_if_exists:
        logging.info(
            f"Found {num_files} NEON data files totaling approximately {download_size}.\n"
            "Files in savepath will be checked and skipped if they exist and match the latest version."
        )
    else:
        logging.info(
            f"Downloading {num_files} NEON data files totaling approximately {download_size}\n"
        )
        # verbose option to list all files being downloaded
        if verbose:
            logging.info("Downloading the following data and metadata files:")
            for file in files:
                logging.info(file.replace("https://storage.googleapis.com", f"{dpid}"))

    if skip_if_exists:
        # Existence and checksum check
        file_url_df[["exists_locally", "checksum_matches"]] = file_url_df.apply(
            lambda row: check_exists_and_checksum(row, download_path), axis=1
        )

        # Handle the various cases
        # 1. Skip files that exist locally and checksums match those on the GCS (exists_locally and checksum_matches are both True).
        # 2. Prompt the user to decide whether to download (overwrite) files that exist locally but checksums don't match (exists_locally is True and checksum_matches is False).
        # 3. Download files if they don't already exist locally (exists_locally is False).
        # 4. If there are extra files locally that don't exist on the GCS, display a warning message.

        # Skipped files
        files_to_skip = file_url_df[
            (file_url_df["exists_locally"]) & (file_url_df["checksum_matches"])
        ]

        # Files to prompt for overwrite
        mismatched_files = file_url_df[
            (file_url_df["exists_locally"]) & (~file_url_df["checksum_matches"])
        ]

        # Files to download (do not exist locally)
        files_to_download = file_url_df[~file_url_df["exists_locally"]]

        # Identify README files (case-insensitive, .txt)
        readme_files = file_url_df[
            file_url_df["name"].str.contains("readme", case=False, na=False)
        ]

        # If any files are missing or mismatched, download the README file
        if (
            not files_to_download.empty or not mismatched_files.empty
        ) and not readme_files.empty:
            logging.info("Downloading README file")
            for idx, row in tqdm(readme_files.iterrows(), total=len(readme_files)):
                download_file(
                    url=row["url"],
                    savepath=download_path,
                    chunk_size=chunk_size,
                    token=token,
                )

        # display a warning message if there are extra files locally that
        # are not published on the Portal as part of the Data Product

        # get all expected file paths (relative to download_path)
        expected_files = set()
        for _, row in file_url_df.iterrows():
            pathparts = row["url"].split("/")
            expected_files.add(
                os.path.normpath(os.path.join(download_path, *pathparts[3:]))
            )

        # get the set of all AOP bucket names from the URLs
        gcs_bucket = {
            url.split("/")[3]
            for url in file_url_df["url"]
            if len(url.split("/")) > 3
            and url.split("/")[3]
            in ["neon-aop-products", "neon-aop-provisional-products"]
        }

        # get the domain from the URLs
        domain = {
            val
            for url in file_url_df["url"]
            for val in [url.split("/")[6]]
            if re.fullmatch(r"D\d{2}", val)
        }

        # get the Year_Site_Visit from the URLs
        ysv = {
            val
            for url in file_url_df["url"]
            for val in [url.split("/")[7]]
            if re.fullmatch(r"\d{4}_[A-Z]{4}_\d+", val)
        }

        # recursively find all files in the expected path
        local_path = os.path.normpath(
            os.path.join(
                download_path,
                gcs_bucket.pop(),
                str(year),
                "FullSite",
                domain.pop(),
                ysv.pop(),
            )
        )

        # get the set of all local files (normalized paths)
        all_local_files = set(
            os.path.normpath(f)
            for f in glob.glob(os.path.join(local_path, "**"), recursive=True)
            if os.path.isfile(f)
        )

        # find extra files and display warning message if any exist
        extra_files = sorted(all_local_files - expected_files)
        if extra_files:
            logging.info(
                f"WARNING: Files found in the local folder that do not exist in the latest version of data for {dpid}. "
                f"\nYou will need to delete the extra files below if you want the download folder to match the latest available data contents,"
                f"\nor to be safe, delete the folder {local_path} and re-download."
                "\nExtra Files Found:"
            )
            for f in sorted(extra_files):
                logging.info(f"  {os.path.abspath(f)}")

        # files that do not exist locally
        if not files_to_download.empty:
            logging.info(
                "The following files will be downloaded (they do not already exist locally):"
            )
            for f in files_to_download["name"].sort_values():
                logging.info(f"  {f}")

            # Download files that do not exist locally
            for _, row in tqdm(
                files_to_download.iterrows(), total=len(files_to_download)
            ):
                download_file(
                    url=row["url"],
                    savepath=download_path,
                    chunk_size=chunk_size,
                    token=token,
                )

            if not files_to_skip.empty:
                logging.info(
                    "The remainder of the files in savepath will not be downloaded. "
                    "They already exist locally and match the latest available data."
                )

        # mismatched files (local checksums are not the same as those on the Data Portal)
        # filter out files where the name contains "readme" (case-insensitive)
        # readme files do not have a checksum so would always fail, these should be downloaded by default
        mismatched_files_no_readme = mismatched_files[
            ~mismatched_files["name"].str.contains("readme", case=False, na=False)
        ]

        # message if there is nothing to download - all files exist and they all match the checksums
        if files_to_download.empty and mismatched_files_no_readme.empty:
            logging.info(
                "All files already exist locally and match the latest available data. Skipping download."
            )

        if not mismatched_files_no_readme.empty:
            logging.info(
                "The following files exist locally but have a different checksum than the remote files:"
            )
            for f in sorted(mismatched_files_no_readme["name"]):
                logging.info(f"  {f}")
            # determine whether to overwrite mismatched files
            if overwrite == "yes":
                response = "y"
            elif overwrite == "no":
                response = "n"
                # logging.info("WARNING: Files where the checksum doesn't match the latest version will not be overwritten.")
            else:  # 'prompt' or any other value
                response = (
                    input(
                        "Do you want to overwrite these files with the latest version? (y/n) "
                    )
                    .strip()
                    .lower()
                )
            if response.lower() == "y":
                logging.info("Overwriting these files with the latest available data.")
                # download including the readme
                for _, row in tqdm(
                    mismatched_files.iterrows(), total=len(mismatched_files)
                ):
                    download_file(
                        url=row["url"],
                        savepath=download_path,
                        chunk_size=chunk_size,
                        token=token,
                    )
            else:
                logging.info("Skipped overwriting files with mismatched checksums.")

    else:
        for file in tqdm(files):
            download_file(
                url=file, savepath=download_path, chunk_size=chunk_size, token=token
            )

    # download issue log table
    ilog = get_issue_log(dpid=dpid, token=None)
    if ilog is not None:
        ilog.to_csv(f"{download_path}/issueLog_{dpid}.csv", index=False)

    # download citations
    if "PROVISIONAL" in releases:
        try:
            cit = get_citation(dpid=dpid, release="PROVISIONAL")
            with open(
                f"{download_path}/citation_{dpid}_PROVISIONAL.txt",
                mode="w+",
                encoding="utf-8",
            ) as f:
                f.write(cit)
        except Exception:
            pass

    rr = re.compile("RELEASE")
    rel = [r for r in releases if rr.search(r)]
    if len(rel) == 0:
        releases = releases
    if len(rel) == 1:
        try:
            cit = get_citation(dpid=dpid, release=rel[0])
            with open(
                f"{download_path}/citation_{dpid}_{rel[0]}.txt",
                mode="w+",
                encoding="utf-8",
            ) as f:
                f.write(cit)
        except Exception:
            pass

    return


# %%


def by_tile_aop(
    dpid,
    site,
    year,
    easting,
    northing,
    buffer=0,
    include_provisional=False,
    check_size=True,
    savepath=None,
    chunk_size=1024,
    token=None,
    verbose=False,
    skip_if_exists=False,
    overwrite="prompt",
):
    """
    This function queries the NEON API for AOP data by site, year, product, and
    UTM coordinates, and downloads all files found, preserving the original
    folder structure. It downloads files serially to avoid API rate-limit
    overload, which may take a long time.

    Parameters
    --------
    dpid: str
        The identifier of the NEON data product to pull, in the form DPL.PRNUM.REV, e.g. DP3.30001.001.

    site: str
        The four-letter code of a single NEON site, e.g. 'CLBJ'.

    year: str or int
        The four-digit year of data collection.

    easting: float/int or list of float/int
        A number or list containing the easting UTM coordinate(s) of the locations to download.

    northing: float/int or list of float/int
        A number or list containing the northing UTM coordinate(s) of the locations to download.

    buffer: float/int, optional
        Size, in meters, of the buffer to be included around the coordinates when determining which tiles to download. Defaults to 0.

    include_provisional: bool, optional
        Should provisional data be downloaded? Defaults to False. See
        https://www.neonscience.org/data-samples/data-management/data-revisions-releases
        for details on the difference between provisional and released data.

    check_size: bool, optional
        Should the user approve the total file size before downloading? Defaults to True.
        If you have sufficient storage space on your local drive, when working
        in batch mode, or other non-interactive workflow, use check_size=False.

    savepath: str, optional
        The file path to download to. Defaults to None, in which case the working directory is used.
        Files are downloaded to subdirectories starting with the dpid under the savepath.

    chunk_size: int, optional
        Size in bytes of chunk for chunked download. Defaults to 1024.

    token: str, optional
        User-specific API token from data.neonscience.org user account. Defaults to None.
        See https://data.neonscience.org/data-api/rate-limiting/ for details about
        API rate limits and user tokens.

    verbose: bool, optional
        If set to True, the function will print out a list of the tiles to be downloaded before downloading.
        Defaults to False.

    skip_if_exists: bool, optional
        If set to True, the function will skip downloading files that already exist in the
        savepath and are valid (local checksums match the checksums of the published file).
        Defaults to False. If any local file checksums don't match those of files published
        on the NEON Data Portal, the user will be prompted to skip these files or overwrite
        the existing files with the new ones (see overwrite input).

    overwrite: str, optional
        Must be one of:
            'yes'    - overwrite mismatched files without prompting,
            'no'     - don't overwrite mismatched files (skip them, no prompt),
            'prompt' - prompt the user (y/n) to overwrite mismatched files after displaying them (default).
        If skip_if_exists is False, this parameter is ignored, and any existing files in
        the savepath will be overwritten according to the function's default behavior.

    Returns
    --------
    None; data are downloaded to the directory specified (savepath) or the current working directory.
    If data already exist in the expected path, they will be overwritten by default. To check for
    existing files before downloading, set skip_if_exists=True along with an overwrite option (y/n/prompt).

    Example
    --------
    >>> by_tile_aop(dpid="DP3.30015.001",
                    site="MCRA",
                    easting=[566456, 566639],
                    northing=[4900783, 4901094],
                    year="2021",
                    savepath="./test_download",
                    skip_if_exists=True)
    # This downloads any tiles overlapping the specified UTM coordinates for
    # 2021 canopy height model data from McRae Creek to the './test_download' directory.
    # If any files already exist in the savepath, they will be checked and skipped if they are valid.
    # The user will be prompted to ovewrite or skip downloading any existing files that do not match
    # the latest published data on the NEON Data Portal.

    Notes
    --------
    This function creates a folder named by the Data Product ID (DPID; e.g. DP3.30015.001) in the
    'savepath' directory, containing all AOP files meeting the query criteria. If 'savepath' is
    not provided, data are downloaded to the working directory, in a folder named by the DPID.
    """

    # raise value error and print message if dpid isn't formatted as expected
    validate_dpid(dpid)

    # raise value error and print message if dpid isn't a valid AOP L3 product
    validate_aop_l3_dpid(dpid)

    # raise value error and print message if field spectra data are attempted
    check_field_spectra_dpid(dpid)

    # raise value error and print message if site is not a 4-letter character
    site = site.upper()  # make site upper case (if it's not already)
    validate_site_format(site)

    # raise value error and print message if site is not a valid NEON site
    validate_neon_site(site)

    # raise value error and print message if year input is not valid
    year = str(year)  # cast year to string (if it's not already)
    validate_year(year)

    # raise value error and print message if skip_if_exists input is not valid (boolean)
    validate_skip_if_exists(skip_if_exists)

    # raise value error and print message if validate_overwrite input is not valid (yes, no, prompt)
    validate_overwrite(overwrite)

    # warn if overwrite is set to yes or no, but skip_if_exists is False
    if not skip_if_exists and overwrite != "prompt":
        logging.info(
            "WARNING: overwrite option only applies if skip_if_exists=True. By default, "
            "any existing files in the expected directory will be overwritten unless "
            "you select skip_if_exists=True and overwrite='no' or 'prompt' (default)."
        )

    # convert easting and northing to lists, if they are not already
    if type(easting) is not list:
        easting = [easting]
    if type(northing) is not list:
        northing = [northing]

    # convert to floats, and display error message if easting and northing lists are not numeric
    try:
        easting = [float(e) for e in easting]
    except ValueError as e:
        logging.info(
            "The easting is invalid, this is required as a number or numeric list format, eg. 732000 or [732000, 733000]"
        )
        print(e)

    try:
        northing = [float(e) for e in northing]
    except ValueError as e:
        logging.info(
            "The northing is invalid, this is required as a number or numeric list format, eg. 4713000 or [4713000, 4714000]"
        )
        print(e)

    # link easting and northing coordinates

    # error message if easting and northing vector lengths don't match (also handles empty/NA cases)
    # there should not be any strings now that everything has been converted to a float
    easting = [e for e in easting if not np.isnan(e)]
    northing = [n for n in northing if not np.isnan(n)]

    if len(easting) != len(northing):
        logging.info(
            "Easting and northing list lengths do not match, and/or contain null values. Cannot identify paired coordinates."
        )
        return

    # if token is an empty string, set to None
    if token == "":
        token = None

    # check for expired token
    if token is not None:
        token = token_check(token)

    # query the products endpoint for the product requested
    response = get_api("https://data.neonscience.org/api/v0/products/" + dpid, token)

    # exit function if response is None (eg. if no internet connection)
    if response is None:
        logging.info("No response from NEON API. Check internet connection")
        return

    # check that token was used
    if token and "x-ratelimit-limit" in response.headers:
        check_token(response)

    # get the request response dictionary
    response_dict = response.json()

    # error message if dpid is not an AOP data product
    if response_dict["data"]["productScienceTeamAbbr"] != "AOP":
        logging.info(
            f"NEON {dpid} is not a remote sensing product. Use zips_by_product()"
        )
        return

    # replace collocated site with the site name it's published under
    site = get_shared_flights(site)

    # get the urls for months with data available, and subset to site & year
    site_year_urls = get_site_year_urls(response_dict, site, year)

    # error message if nothing is available
    if len(site_year_urls) == 0:
        logging.info(
            f"There are no NEON {dpid} data available at the site {site} in {year}.\nTo display available dates for a given data product and site, use the function list_available_dates()."
        )
        return

    # get file url dataframe for the available month url(s)
    file_url_df, releases = get_file_urls(site_year_urls, token=token)

    # get the number of files in the dataframe, if there are no files to download, return
    if len(file_url_df) == 0:
        logging.info("No NEON data files found.")
        return

    # if 'PROVISIONAL' in releases and not include_provisional:
    if include_provisional:
        # print provisional included message
        logging.info(
            "Provisional NEON data are included. To exclude provisional data, use input parameter include_provisional=False."
        )
    else:
        # print provisional not included message
        file_url_df = file_url_df[file_url_df["release"] != "PROVISIONAL"]
        logging.info(
            "Provisional NEON data are not included. To download provisional data, use input parameter include_provisional=True."
        )

        # get the number of files in the dataframe after filtering for provisional data, if there are no files to download, return
        num_files = len(file_url_df)
        if num_files == 0:
            logging.info(
                "No NEON data files found. Available data may all be provisional. To download provisional data, use input parameter include_provisional=True."
            )
            return

    # BLAN edge-case - contains plots in 18N and plots in 17N; flight data are all in 17N
    # convert easting & northing coordinates for Blandy (BLAN) to 17N

    if site == "BLAN" and any([e <= 250000.0 for e in easting]):
        # print('BLAN SITE - CONVERTING COORDINATES FROM 18N TO 17N')
        # check that pyproj is installed
        try:
            from pyproj import Proj, CRS
        except ImportError:
            logging.info(
                "Package pyproj is required for this function to work at the NEON BLAN site. Install and re-try."
            )
            return

        crs17 = CRS.from_epsg(32617)  # utm zone 17N
        crs18 = CRS.from_epsg(32618)  # utm zone 18N

        proj18to17 = Proj.from_crs(crs_from=crs18, crs_to=crs17)

        # link easting and northing coordinates so it's easier to parse the zone for each
        coord_tuples = [(easting[i], northing[i]) for i in range(0, len(easting))]

        coords17 = [(e, n) for (e, n) in coord_tuples if e > 250000.0]
        coords18 = [(e, n) for (e, n) in coord_tuples if e <= 250000.0]

        # apply the projection transformation from 18N to 17N for each coordinate tuple

        coords18_reprojected = [
            proj18to17.transform(coords18[i][0], coords18[i][1])
            for i in range(len(coords18))
        ]

        coords17.extend(coords18_reprojected)

        # re-set easting and northing
        easting = [c[0] for c in coords17]
        northing = [c[1] for c in coords17]

        logging.info(
            "Blandy (BLAN) plots include two UTM zones, flight data "
            "are all in 17N. Coordinates in UTM zone 18N have been "
            "converted to 17N to download the correct tiles. You "
            "will need to make the same conversion to connect "
            "airborne to ground data."
        )

    # function to round down to the nearest 1000, in order to determine
    # lower left coordinate of AOP tile to be downloaded
    def round_down1000(val):
        return int(np.floor(val / 1000) * 1000)

    # function to get the coordinates of the tiles including the buffer
    def get_buffer_coords(easting, northing, buffer):
        # apply the buffer to the easting and northings
        buffer_min_e = easting - buffer
        buffer_min_n = northing - buffer
        buffer_max_e = easting + buffer
        buffer_max_n = northing + buffer

        new_coords = [
            (buffer_min_e, buffer_min_n),
            (buffer_min_e, buffer_max_n),
            (buffer_max_e, buffer_min_n),
            (buffer_max_e, buffer_max_n),
        ]

        return new_coords

    # get the tiles corresponding to the new coordinates (mins and maxes)
    buffer_coords = []
    for e, n in zip(easting, northing):
        buffer_coords.extend(get_buffer_coords(e, n, buffer))

    buffer_coords_rounded = [
        (round_down1000(c[0]), round_down1000(c[1])) for c in buffer_coords
    ]
    # remove duplicate coordinates
    buffer_coords_set = list(set(buffer_coords_rounded))
    buffer_coords_set.sort()

    utm17_eastings_str = ", ".join([str(round(e, 2)) for e in easting])
    utm17_northings_str = ", ".join(str(round(n, 2)) for n in northing)

    if site == "BLAN" and verbose:
        logging.info(f"UTM 17N Easting(s): {utm17_eastings_str}")
        logging.info(f"UTM 17N Northing(s): {utm17_northings_str}")
        # logging.info('Buffer:', buffer)
    if verbose:
        logging.info("UTM (x, y) lower-left coordinates of tiles to be downloaded:")
        for coord in buffer_coords_set:
            logging.info(coord)

    # create the list of utm "easting_northing" strings that will be used to match to the tile names
    coord_strs = ["_".join([str(c[0]), str(c[1])]) for c in buffer_coords_set]

    # append the .txt file to include the README - IS THIS NEEDED?
    coord_strs.append(".txt")

    # subset the dataframe to include only the coordinate strings matching coord_strs
    file_url_df_subset = file_url_df[
        file_url_df["name"].str.contains("|".join(coord_strs))
    ]

    # remove .txt files (README) and create a copy to use for the checksum/existence checks
    file_url_df_subset2 = file_url_df_subset[
        ~file_url_df_subset["name"].str.endswith(".txt")
    ].copy()

    # file_url_df_subset2.reset_index(drop=True,inplace=True)
    file_url_df_subset2 = file_url_df_subset2.reset_index(drop=True)

    # print(list(file_url_df_subset2.head(1)))
    # print(file_url_df_subset2.head())

    # if any coordinates were not included in the data, print a warning message
    unique_coords_to_download = set(
        file_url_df_subset2["name"].str.extract(r"_(\d{6}_\d{7})_")[0]
    )

    coord_strs.remove(".txt")
    # print('coordinates:'); print(coord_strs)
    # print('unique coordinates to download:'); print(unique_coords_to_download)
    # find the coordinates that were not found in the data (difference between the two lists):
    coords_not_found = list(set(coord_strs).difference(list(unique_coords_to_download)))
    if len(coords_not_found) > 0:
        logging.info(
            "WARNING: the following coordinates fall outside the bounds of the NEON site, so will not be downloaded:"
        )
        for coord in coords_not_found:
            print(",".join(coord.split("_")))

    # get the number of files in the dataframe, if there are no files to download, return
    num_files = len(file_url_df_subset)
    if num_files == 0:
        logging.info(f"WARNING: No NEON {dpid} files found.")
        return

    # get the total size of all the files found
    download_size_bytes = file_url_df_subset["size"].sum()

    # convert to a human-readable format
    download_size = convert_byte_size(download_size_bytes)

    # ask whether to continue download, depending on size
    if check_size:
        if (
            input(
                f"Continuing will download {num_files} NEON data files totaling approximately {download_size}. Do you want to proceed? (y/n) "
            )
            .strip()
            .lower()
            != "y"
        ):
            print("Download halted")
            return

    # create folder in working directory to put files in
    if savepath is not None:
        download_path = savepath + "/" + dpid
    else:
        download_path = os.getcwd() + "/" + dpid
    # print('download path', download_path)
    os.makedirs(download_path, exist_ok=True)

    # Get the sorted list of the files to download
    # use the files from the subsetted dataframe
    files = list(file_url_df_subset["url"])
    files.sort()  # sort the files for consistent download order

    # different messages depending on whether skip_if_exists is True or False
    if skip_if_exists:
        logging.info(
            f"Found {num_files} NEON data files totaling approximately {download_size}.\n"
            "Files in savepath will be checked and skipped if they exist and match the latest version."
        )
    else:
        logging.info(
            f"Downloading {num_files} NEON data files totaling approximately {download_size}\n"
        )
        if verbose:
            logging.info("Downloading the following data and metadata files:")
            for file in files:
                logging.info(file.replace("https://storage.googleapis.com", f"{dpid}"))

    if skip_if_exists:
        # Check which files already exist locally and if their checksums match
        file_url_df_subset2[
            ["exists_locally", "checksum_matches"]
        ] = file_url_df_subset2.apply(
            lambda row: check_exists_and_checksum(row, download_path), axis=1
        )

        # Files to skip (already exist locally and checksums match)
        files_to_skip = file_url_df_subset2[
            (file_url_df_subset2["exists_locally"])
            & (file_url_df_subset2["checksum_matches"])
        ]

        # Files to prompt for overwrite (exists locally but checksums don't match)
        mismatched_files = file_url_df_subset2[
            (file_url_df_subset2["exists_locally"])
            & (~file_url_df_subset2["checksum_matches"])
        ]

        # Files to download (do not exist locally)
        files_to_download = file_url_df_subset2[~file_url_df_subset2["exists_locally"]]
        # print(file_url_df_subset2[["name","exists_locally","checksum_matches"]])

        # if verbose: # print these even in non-verbose mode
        if not files_to_download.empty:
            logging.info(
                "The following files will be downloaded (they do not already exist locally):"
            )
            for f in files_to_download["name"].sort_values():
                logging.info(f"  {f}")

            # Download files that do not exist locally
            for _, row in tqdm(
                files_to_download.iterrows(), total=len(files_to_download)
            ):
                download_file(
                    url=row["url"],
                    savepath=download_path,
                    chunk_size=chunk_size,
                    token=token,
                )

            if not files_to_skip.empty:
                logging.info(
                    "The remainder of the files in savepath will not be downloaded. "
                    "They already exist locally and match the latest available data."
                )

        # prompt for mismatched files (excluding readme.txt)
        mismatched_files_no_readme = mismatched_files[
            ~mismatched_files["name"].str.contains("readme", case=False, na=False)
        ]

        # identify README files (case-insensitive, usually .txt), there should only be one of these
        readme_files = mismatched_files[
            mismatched_files["name"].str.contains("readme", case=False, na=False)
        ]

        # download README files
        if not readme_files.empty:
            logging.info("Downloading README file")
            for _, row in tqdm(readme_files.iterrows(), total=len(readme_files)):
                download_file(
                    url=row["url"],
                    savepath=download_path,
                    chunk_size=chunk_size,
                    token=token,
                )

        # message if there is nothing to download - all files exist and they all match the checksums
        if files_to_download.empty and mismatched_files_no_readme.empty:
            logging.info(
                "All files already exist locally and match the latest available data. Skipping download."
            )

        if not mismatched_files_no_readme.empty:
            logging.info(
                "The following files exist locally but have a different checksum than the remote files:"
            )
            for f in sorted(mismatched_files_no_readme["name"]):
                logging.info(f"  {f}")
            # determine whether to overwrite mismatched files
            if overwrite == "yes":
                response = "y"
            elif overwrite == "no":
                response = "n"
            else:  # 'prompt' or any other value
                response = (
                    input(
                        "Do you want to overwrite these files with the latest version? (y/n) "
                    )
                    .strip()
                    .lower()
                )

            if response.lower() == "y":
                logging.info("Overwriting these files with the latest available data.")
                for idx, row in tqdm(
                    mismatched_files.iterrows(), total=len(mismatched_files)
                ):
                    download_file(
                        url=row["url"],
                        savepath=download_path,
                        chunk_size=chunk_size,
                        token=token,
                    )
            else:
                logging.info("Skipped overwriting files with mismatched checksums.")
    else:  # if skip_if_exists=False (default behavior)
        for file in tqdm(files):
            download_file(
                url=file, savepath=download_path, chunk_size=chunk_size, token=token
            )

    # download issue log table
    ilog = get_issue_log(dpid=dpid, token=None)
    if ilog is not None:
        ilog.to_csv(f"{download_path}/issueLog_{dpid}.csv", index=False)

    # download citations
    if "PROVISIONAL" in releases:
        try:
            cit = get_citation(dpid=dpid, release="PROVISIONAL")
            with open(
                f"{download_path}/citation_{dpid}_PROVISIONAL.txt",
                mode="w+",
                encoding="utf-8",
            ) as f:
                f.write(cit)
        except Exception:
            pass

    rr = re.compile("RELEASE")
    rel = [r for r in releases if rr.search(r)]
    if len(rel) == 0:
        releases = releases
    if len(rel) == 1:
        try:
            cit = get_citation(dpid=dpid, release=rel[0])
            with open(
                f"{download_path}/citation_{dpid}_{rel[0]}.txt",
                mode="w+",
                encoding="utf-8",
            ) as f:
                f.write(cit)
        except Exception:
            pass

    return



# -*- coding: utf-8 -*-
"""
Created on 15 Aug 2024

@author: Claire Lunch

Unit tests for load_by_product()

"""

# import required packages
from src.neonutilities.unzip_and_stack import load_by_product
import pandas as pd


def test_load_by_product_IS():
    """
    Test that load_by_product() accesses and stacks the example data correctly.
    """
    tlist = load_by_product(dpid='DP1.00005.001', site=['TOOL', 'PUUM'],
                            startdate='2022-06', enddate='2022-07',
                            check_size=False, progress=False,
                            release='RELEASE-2024', cloud_mode=True)
    assert list(tlist.keys()) == ['IRBT_1_minute', 'IRBT_30_minute', 'citation_00005_RELEASE-2024',
                                  'issueLog_00005', 'readme_00005', 'science_review_flags_00005',
                                  'sensor_positions_00005', 'variables_00005']
    tm = tlist['IRBT_30_minute']
    assert len(tm) == 23424
    assert max(tm['endDateTime']) == pd.Timestamp('2022-08-01 00:00:00+0000', tz='UTC')
    assert min(tm['endDateTime']) == pd.Timestamp('2022-06-01 00:30:00+0000', tz='UTC')



import pandas as pd
import os
import zipfile
from datetime import datetime
import re
import gzip
import shutil
import h5py
from tqdm import tqdm
import numpy as np
import warnings
import logging
from .citation import get_citation
from .get_issue_log import get_eddy_issue_log


def list_h5_contents(file_path):
    """
    List the contents of an H5 data file
    
    Parameters
    --------
        file_path (str): Path to the H5 file

    Return
    --------
        A data frame of H5 file contents
        
    """
    contents_ = []
    
    with h5py.File(file_path, 'r') as hdf:
        def create_listObj(group, otype):
            if group.count('/') == 0:
                name = group
                group = '/'
            else:
                groupSplit = group.split('/')
                name = groupSplit[-1]
                group = '/' + '/'.join(groupSplit[0:-1])
            
            contents_.append({'group': group, 'name': name, 'otype': type(otype).__name__})
        hdf.visititems(create_listObj)

    listObj = pd.DataFrame(contents_)
    return listObj


def time_stamp_set(tabDict):
    """
    Generate consensus SAE time stamps from a set of tables. Used in stack_eddy(), not intended for independent use.

    Parameters
    --------
        tabDict: A dictionary of SAE data frames

    Return
    --------
        A data frame of time stamps (start and end times) aggregated from the input tables
    
    """
    nameSet = ['timeBgn', 'timeEnd']
    
    # get a set of time stamps to initiate the table. leave out qfqm to exclude 
    #   - filler records created as placeholders for days with no data
    #   - turbulent flux and footprint end time stamps don't quite match the others
    timeSet = {}
    for key in tabDict.keys():
        if not re.search('qfqm|turb|foot', key):
            timeSet[key] = tabDict[key].copy()
    
    # initiate the table with consensus set of time stamps
    timeSetInit = timeSet[list(timeSet.keys())[0]][nameSet]
    
    if len(timeSet.keys()) == 1:
        pass
    else:
        for q in list(timeSet.keys())[1:]:
            # check for additional start time stamps
            timeSetTemp = timeSet[q][nameSet]
            timeSetTempMerg = pd.DataFrame(timeSetTemp['timeBgn'])
            timeSetInitMerg = pd.DataFrame(timeSetInit['timeBgn'])
            misTime = timeSetTempMerg.merge(timeSetInitMerg, on = 'timeBgn', how = 'left', indicator = True)
            misTime = misTime[misTime['_merge'] == 'left_only'].drop(columns = ['_merge'])
    
            if len(misTime) == 0:
                pass
            else:
                # combine all, then de-dup
                timeSetInit = pd.concat([timeSetInit, timeSetTemp], axis = 0, ignore_index = True).drop_duplicates()
    return timeSetInit


def eddy_stamp_check(tab):
    """
    Convert SAE time stamps to datetime and check for missing data

    Parameters
    --------
        tab: A data frame of SAE data

    Return
    --------
        The same table of SAE data, with time stamps converted and empty records representing a single day (filler records inserted during processing) removed.
    
    """
    tab = tab.copy()
    # convert time stamps
    tBgnErr = False
    tEndErr = False
    
    try:
        tabBP = pd.to_datetime(tab['timeBgn'].apply(lambda x: x.decode('utf-8')), format="%Y-%m-%dT%H:%M:%S.%fZ").dt.tz_localize('UTC').dt.tz_convert('GMT')
    except: 
        tBgnErr = True
    
    try:
        tabEP = pd.to_datetime(tab['timeEnd'].apply(lambda x: x.decode('utf-8')), format="%Y-%m-%dT%H:%M:%S.%fZ").dt.tz_localize('UTC').dt.tz_convert('GMT')
    except:
        tEndErr = True
    
    # if conversion failed, keep character time stamps and pass along message
    err = False
    
    if tBgnErr:
        err = True
    else:
        tab['timeBgn'] = tabBP
    
    if tEndErr:
        err = True
    else:
        tab['timeEnd'] = tabEP
    
    # if conversion was successful, check for single-day empty records and remove
    if err:
        pass
    else:
        dayDiff = [(EP - BP).total_seconds() for EP, BP in zip(tabEP, tabBP)]
        dayDup = [i for i, diff in enumerate(dayDiff) if diff > 86399]
        if len(dayDup) == 0:
            pass
        else:
            tab.drop(dayDup, inplace = True)

    return [tab, err]


def get_attributes(fil, sit, attType, valName = None):
    """
    Extract attribute metadata from H5 files

    Parameters
    --------
        fil: File path to the H5 file to extract attributes from
        sit: The site, for site attributes. Must match site of file path
        attType: The type of attributes to retrieve
        valName: If CO2 validation metadata are requested, the H5 name of the level where they can be found
    
    Return
    --------
        A dictionary containing the extracted attributes

    """
    if fil.count('basic') >= 1:
        mnth = re.search(r"20[0-9]{2}-[0-9]{2}", fil).group()
    else:
        mnth = re.search(r"20[0-9]{2}-[0-9]{2}-[0-9]{2}", fil).group()
    
    if attType == 'site':
        lev = str(sit)
    if attType == 'root':
        lev = '/'
    if attType == 'val':
        lev = valName
    
    try: 
        with h5py.File(fil, 'r') as hdf:
            gAttr = dict(hdf[lev].attrs)
            for key, val in gAttr.items():
                try:
                    gAttr[key] = [item.decode('utf-8') for item in val]
                except:
                    pass
            gAttr['site'] = [str(sit)]
            gAttr['date'] = [mnth]
    except:
        gAttr = {'site': [str(sit)], 'date': [mnth]}
    
    return gAttr


def get_variables_eddy(tabDict):
    """
    Extract variable names and units from SAE H5 files and return in user-friendly form. Used in stack_eddy(), not intended for independent use.
    
    Parameters
    -------- 
        tabDict: A dictionary of SAE data frames

    Return
    -------- 
        A data frme of variable names and units, aggregated from the input tables
    
    """
    variables = pd.DataFrame({'category': [], 'system': [], 'variable': [], 'stat': [], 'units': []})
    for p in tabDict.keys():
        for q in tabDict[p].keys():
            if 'unit' in tabDict[p][q].keys():
                var_nm = []
                keySplit = q.split('/')
                var_nm.append(keySplit[3])
                var_nm.append(keySplit[4])
                var_nm.append(keySplit[-1])
                units = tabDict[p][q]['unit']
            
                if len(units) > 1:
                    variables_q = pd.DataFrame({'category': [var_nm[0]] * len(units), 'system': [var_nm[1]] * len(units), 'variable': [var_nm[2]] * len(units)})
                    names = tabDict[p][q]['data'].columns
                    
                    if len(units) == len(names):
                        variables_q['stat'] = names
                    else:
                        if 'index' in names:
                            variables_q['stat'] = [name for name in names if name != 'index']
                        else:
                            variables_q['stat'] = [name for name in names if name not in ['timeBgn', 'timeEnd']]
            
                    variables_q['units'] = units
                    variables = pd.concat([variables, variables_q], ignore_index = True)
                else:
                    variables_q = pd.DataFrame({'category': [var_nm[0]], 'system': [var_nm[1]], 'variable': [var_nm[2]], 'stat': None, 'units': units})
                    variables = pd.concat([variables, variables_q], ignore_index = True)
            
    variables.drop_duplicates(inplace = True, ignore_index = True)
    
    if len(variables) == 0:
        variables = None
    
    return variables


def get_vars_eddy(filepath):
    """
    Extracts a data frame of table metadata from a single HDF5 file. Specific to eddy covariance data product DP4.00200.001. Can inform inputs to stack_eddy(); variables listed in 'name' are available inputs to the 'var' parameter in stack_eddy().
    
    Parameters
    --------
        filepath (str): The path to the H5 file
    
    Return
    --------
        A data frame of the metadata for each data table in the HDF5 file
    
    """
    try:
        listObj = list_h5_contents(filepath)
        listDataObj = listObj.loc[listObj['otype'] == 'Dataset']
    
        listObjSpl = listDataObj.copy()
        listObjSpl[['blank', 'site', 'level', 'category', 'system', 'horvertmi']] = listObjSpl['group'].str.split('/', expand = True)
        listObjSpl[['hor', 'ver', 'tmi']] = listObjSpl['horvertmi'].str.split('_', expand = True)
        listObjSpl.drop(columns=['group','blank', 'horvertmi'], axis=1, inplace=True)
    
        return listObjSpl
    except:
        raise ValueError(filepath + ' could not be read.')


def stack_eddy(filepath, level='dp04', var=None, avg=None, metadata=False, runLocal=False):
    """
    Convert data of choice from HDF5 to tabular format. Specific to eddy covariane data product: DP4.00200.001

    Parameters
    --------
        filepath: One of 1) a folder containing NEON EC H5 files, 2) a zip file of DP4.00200.001 data downloaded from the data portal, 3) a list of H5 files, 4) a single EC H5 file
        level: The level of data to extract; one of dp01, dp02, dp03, dp04
        var: The variable set to extract. Can be any of the variables in the "name" level of the "system" level of the H5 file; use getVarsEddy() function to see the available variables. From the inputs, all variables from "name" and all variables from "system" will be returned, but if variables from both "name" and "system" are specified, the function will return only the intersecting set. This allows the user to, e.g., return only the pressure data ("pres") from the CO2 storage system ("co2Stor"), instead of all the pressure data from all instruments.
        avg: The averaging interval to extract, in minutes.
        metadata: Should the output include metadata from the attributes of the H5 files? Defaults to false. Even when false, variable definitions, issue logs, and science review flags will be included.
        runLocal: Set to True to omit any calls to the NEON API. Data are extracted and reformatted from local files, but citation and issue log are not retrieved.

    Return
    --------
        varMergDict: A dictionary of dataframes. One data frame per site, plus one dat frame containing the metadata (objDesc) table and one data frame containing units for each variable.
    
    """

    files = []
    releases = []
    skipKeys = []
    
    # check input types
    if isinstance(var, str):
        var = [var]
    if not isinstance(level, str):
        raise ValueError("level must be one of dp01, dp02, dp03, dp04")
    if not level in ["dp01", "dp02", "dp03", "dp04"]:
        raise ValueError("level must be one of dp01, dp02, dp03, dp04")

    # check for list of h5 files as input
    if isinstance(filepath, list):
        if sum(path.count(".h5") for path in filepath) == len(filepath):
            files = filepath
        else:
            raise ValueError("Input list of files must be .h5 files.")
        for path in filepath:
            if not os.path.exists(path):
                raise ValueError("Files not found in specified filepaths. Check that the input list contains the correct filepath.")
            
    # get list of files, unzipping if necessary
    if len(files) == 0 and isinstance(filepath, str) and filepath.endswith(".zip"):
        outpath = filepath.replace(".zip", "")
        
        os.makedirs(outpath, exist_ok = True)
        
        with zipfile.ZipFile(filepath, 'r') as zip_ref:
            zip_list = zip_ref.namelist()
            
            if any(".zip" in name for name in zip_list):
                zip_ref.extractall(outpath)
            
            else:
                # get release info before discarding file paths
                all_dirs = [os.path.dirname(name) for name in zip_list]
                pattern = '(RELEASE|PROVISIONAL|LATEST).*'
                release_status = []
                for file in all_dirs:
                    match_ = re.finditer(pattern, file)
                    for match in match_:
                        release_status.append({'name': file, 'release': match.group(0)})
                
                release_status = pd.DataFrame(release_status)
                releases = release_status['release'].unique()
                
                # write release status file
                timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
                release_status.to_csv(os.path.join(outpath, f"release_status_{timestamp}.csv"), index=False)
                
                # unzip
                for file in zip_list:
                    zip_ref.extract(file, outpath)
        
        filepath = outpath

    # allow for a single H5 file, or folder of H5 files
    if len(files) == 0 and isinstance(filepath, str) and filepath.endswith(".h5"):
        files = [filepath]
    else: 
        if len(files) == 0:
            for root, dirs, filenames in os.walk(filepath):
                for filename in filenames:
                    files.append(os.path.join(root, filename))

    # unzip files if necessary
    if sum(file.count(".zip") for file in files) == len(files):
        for file in files:
            with zipfile.ZipFile(file, 'r') as zip_ref:
                zip_ref.extractall(filepath)
        files = os.listdir(filepath)

    # after unzipping, check for .gz
    if any(file.endswith(".h5.gz") for file in files):
        gzfiles = [file for file in files if re.search(".h5.gz", file)]
        
        for gzfile in gzfiles:
            with gzip.open(gzfile, 'rb') as f_in:
                with open(gzfile.removesuffix(".gz"), 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
            os.remove(gzfile)

        files = []
        for root, dirs, filenames in os.walk(filepath):
            for filename in filenames:
                files.append(os.path.join(root, filename))

    # need the H5 files for data extraction and the SRF tables
    scienceReviewList = [file for file in files if re.search("science_review_flags", file)]
    files = [file for file in files if re.search(".h5$", file)]

    # check for no files
    if len(files) == 0:
        raise ValueError("No .h5 files found in specified file path. Check the inputs and file contents.")

    # check for original zip files and use to determine releases and citations
    if len(releases) == 0 and isinstance(filepath, str) and os.path.isdir(filepath):
        allFiles = os.listdir(filepath)
        relfl = [file for file in allFiles if re.search("release_status", file)]
        if len(relfl) == 1:
            reltab = pd.read_csv(relfl)
            releases = reltab['release'].unique()
        else:
            allFileNames = [os.path.splitext(file)[0] for file in allFiles]
            pattern = '(RELEASE|PROVISIONAL|LATEST).*'
            release_status = []
            for file in allFileNames:
                match_ = re.finditer(pattern, file)
                for match in match_:
                    release_status.append(match.group(0))
            
            releases = list(set(release_status))

    # get DOIs and generate citations(s)
    if not runLocal:
        citP = None
        citR = None
        if "PROVISIONAL" in releases:
            try:
                citP = get_citation(dpid = "DP4.00200.001", release = "PROVISIONAL")
            except:
                citP = None
        if sum(release.count("RELEASE") for release in releases) == 0:
            pass
        else: 
            if sum(release.count("RELEASE") for release in releases) > 1:
                raise ValueError("Attempting to stack multiple data releases together. This is not appropriate, check your input data")
            else:
                rel = [release for release in releases if re.search("RELEASE", release)]
                try:
                    citR = get_citation(dpid = "DP4.00200.001", release = rel[0])
                except:
                    citR = None

    # determine basic vs expanded package and check for inconsistencies
    pkgCount = sum(bool(re.search("basic\\.[a-zA-Z0-9]+\\.(h5)$", file)) for file in files)

    if len(files) == pkgCount:
        pkg = 'basic'
    elif pkgCount == 0:
        pkgCount = len([re.findall("expanded\\.[a-zA-Z0-9]+\\.h5$", file) for file in files])
        pkg = 'expanded'

    if len(files) != pkgCount:
        raise ValueError("File path contains both basic and expanded package files, these can't be stacked together.")

    if pkg == 'basic' and metadata == True:
        logging.info("For the basic package, attribute metadata are the values from the beginning of the month. To get attributes for each day, use the expanded package.")

    # make empty dictionary for the data tables
    tableDict = {}

    for file in files:
        tableDict[file.replace(".h5", "")] = []

    # extract data from each file
    for file in tqdm(files, desc = "Extracting data"):
        try:
            if isinstance(filepath, list):
                h5path = file
            else:
                h5path = os.path.join(filepath, file)
                
            listObj = list_h5_contents(h5path)
            listDataObj = listObj.loc[listObj['otype'] == 'Dataset'].reset_index()
            listDataName = listDataObj['group'] + '/' + listDataObj['name']
            
            listObjSpl = pd.DataFrame(listDataObj['group'].copy())
            colSpl = listObjSpl['group'].str.split('/', expand = True)
            for i in range(colSpl.shape[1], 7):
                colSpl[i] = None
            colSpl.columns = ['blank', 'site', 'level', 'category', 'system', 'horvertmi', 'subsys']
            listObjSpl = pd.concat([listObjSpl, colSpl], axis=1)
            listObjSpl.drop(columns=['group', 'blank'], axis=1, inplace=True)
            
            # filter by variable/level selections
            levelInd = pd.Index([listDataName.index[i] for i, name in enumerate(listDataName) if re.search(level, name)])
            
            if level not in ["dp04", "dp03", "dp02"] and var is not None:
                if len([value for value in listObjSpl['system'].values if value in var]) > 0:
                    if len([value for value in listDataObj['name'].values if value in var]) > 0:
                        varInd = listObjSpl.loc[listObjSpl['system'].isin(var)].index.intersection(listDataObj.loc[listDataObj['name'].isin(var)].index)
                    else:
                        varInd = listObjSpl.loc[listObjSpl['system'].isin(var)].index
                else: 
                    if len([value for value in listDataObj['name'].values if value in var]) > 0:
                        varInd = listDataObj.loc[listDataObj['name'].isin(var)].index
                    else:
                        raise ValueError("No data found for variables: " + ", ".join(var))
            else:
                varInd = pd.Index(range(0, len(listDataName)))
            
            if level not in ["dp04", "dp03", "dp02"] and avg is not None:
                avgInd = pd.Index([listDataName.index[i] for i, name in enumerate(listDataName) if re.search(str(avg) + 'm', name)])
            else:
                if level == "dp01" and avg is None:
                    raise ValueError("If level is dp01, avg is a required input.")
                else:
                    avgInd = pd.Index(range(0, len(listDataName)))
            
            # exclude footprint grid data from expanded packages
            if sum(name.count("foot/grid") for name in listDataName) > 0:
                gridInd = pd.Index([listDataName.index[i] for i, name in enumerate(listDataName) if not re.search('foot/grid', name)])
            else:
                gridInd = pd.Index(range(0, len(listDataName)))
            
            # index that includes all filtering criteria
            ind = gridInd.intersection(levelInd).intersection(varInd).intersection(avgInd)
            
            # check that you haven't filtered to nothing
            tableDictKey = file.replace(".h5", "")
            if len(ind) == 0:
                tableDict["skip" + str(i)] = tableDict.pop(tableDictKey, None)
                skipKeys.append("skip" + str(i))
                logging.info("There are no data meeting the criteria level " + str(level) + ", averaging interval " + str(avg) + ", and variables " + ", ".join(var))
            else:
                listDataName = listDataName[ind]
                
                # add extracted data to the dictionary
                dataListDict = {}
                with h5py.File(h5path, 'r') as hdf:
                    for dataSet in listDataName:
                        dataListDict[dataSet] = {}
                        dataListDict[dataSet]['data'] = pd.DataFrame(hdf[dataSet][:])
                        try:
                            dataListDict[dataSet]['unit'] = [b.decode('utf-8') for b in list(hdf[dataSet].attrs['unit'])]
                        except:
                            pass
                        
                tableDict[tableDictKey] = dataListDict
        except ValueError:
            logging.info('An input value was invalid.')
            raise
        except Exception:
            logging.info(file + " could not be read.")
            pass
        else:
            pass

    # remove skipped files from dict list
    for key in skipKeys:
        tableDict.pop(key, None)

    # check for no data left
    if len(tableDict) == 0:
        raise ValueError("No data met requested criteria in any file")
    
    # get variable units
    variables = get_variables_eddy(tableDict)

    # convert all time stamps to time format, then filter out instances with:
    # 1) only one record for a day
    # 2) all values = NaN
    # these are instances when a sensor was offline, and they don't join correctly
    err = False

    for k1 in tableDict.keys():
        for k2 in tableDict[k1].keys():
            tabTemp = eddy_stamp_check(tableDict[k1][k2]['data'])
            if tabTemp[1]:
                err = True
            tableDict[k1][k2]['data'] = tabTemp[0]

    if err:
        logging.info("Some time stamps could not be converted. Variable join may be affected; check data carefully for disjointed time stamps")

    # within each site-month set, join matching tables
    # create empty dict for the tables
    mergTableDict = {}
    for key in tableDict.keys():
        mergTableDict[key] = {}

    for key in tqdm(tableDict.keys(), desc = "Joining data variables by file"):
        namesSpl = []
        for key2 in tableDict[key].keys():
            namesSpl.append(key2.split('/')[1:])
        
        namesSpl = pd.DataFrame(namesSpl)
        nc = len(namesSpl.columns)
        
        # dp01 and dp02 have sensor levels
        if nc == 6:
            sens = namesSpl[4].unique()
        
            # join for each sensor level
            for si in sens:
                # get all tables for a sensor level and get variable names
                tbsub = {}
                nmsub = []
                inds = namesSpl.loc[namesSpl[4] == si, :]
                for ind, row in inds.iterrows():
                    tkey = '/' + '/'.join(row)
                    tbsub[tkey] = tableDict[key][tkey]['data'].copy()
                    nmsub.append('.'.join([row[2], row[3], row[5]]))
                    mergTabl = time_stamp_set(tbsub)
        
                for i, ib in enumerate(tbsub.keys()):
                    tbsub[ib].rename(columns = {col: (nmsub[i] + '.' + col if 'timeBgn' not in col and 'timeEnd' not in col else col) for col in tbsub[ib].columns}, inplace = True)
                    tbsub[ib].drop(columns = 'timeEnd', inplace = True)
        
                    mergTabl = pd.merge(mergTabl, tbsub[ib], on = 'timeBgn', how = 'left')
        
                mergTableDict[key][si] = mergTabl

        # dp03 and dp04 - no sensor levels
        # get consensus time stamps
        else:
            tbsub = {}
            nmsub = []
            for ind, row in namesSpl.iterrows():
                tkey = '/' + '/'.join(row)
                tbsub[tkey] = tableDict[key][tkey]['data'].copy()
                nmsub.append('.'.join([row[2], row[3], row[4]]))
                mergTabl = time_stamp_set(tbsub)
            
            for i, ibl in enumerate(tbsub.keys()):
                tbsub[ibl].rename(columns = {col: (nmsub[i] + '.' + col if 'timeBgn' not in col and 'timeEnd' not in col else col) for col in tbsub[ibl].columns}, inplace = True)
                tbsub[ibl].drop(columns = 'timeEnd', inplace = True)
            
                mergTabl = pd.merge(mergTabl, tbsub[ibl], on = 'timeBgn', how = 'left')
            
                mergTableDict[key] = mergTabl

    # for dp01 and dp02, stack tower levels and calibration gases
    if level in ['dp01', 'dp02']:
        verMergDict = {}
        for ni in mergTableDict.keys():
            verMergList = []
            for mi in mergTableDict[ni].keys():
                verSpl = mi.split('_')
                
                if len(verSpl) == 3:
                    mergTableDict[ni][mi].insert(0, 'verticalPosition', verSpl[1])
                    mergTableDict[ni][mi].insert(0, 'horizontalPosition', verSpl[0])
                else:
                    mergTableDict[ni][mi].insert(0, 'verticalPosition', verSpl[0])
                    mergTableDict[ni][mi].insert(0, 'horizontalPosition', None)

                verMergList.append(mergTableDict[ni][mi])
            verMergDict[ni] = pd.concat(verMergList, ignore_index = True)

    else:
        verMergDict = mergTableDict.copy()

    # check for weird isotope joining - return to this with a avg = 9 example
    if level == 'dp01' and avg in [9, '9', '09']:
        allNm = []
        for key in verMergDict.keys():
            allNm.append(verMergDict[key].columns)
        allNm = np.unique(allNm)

        if sum(col.count('dlta13CCo2') for col in allNm) > 0 and sum(col.count('dlta18OH2o') for col in allNm) > 0 or sum(col.count('dlta2HH2o') for col in allNm) > 0:
            dup_iso_list = [(x['timeEnd'].iloc[:20] - x['timeBgn'].iloc[:20]).dt.total_seconds().mean(skipna=True) < 537 for x in verMergList]
            dup_iso = any(dup_iso_list)

            if dup_iso:
                warnings.warn("Stacking appears to include both carbon and water isotopes, with inconsistent time stamps. "
                            "Carbon isotopes are measured every 6 minutes, water isotopes every 9 minutes. "
                            "This issue affects RELEASE-2023 and provisional data published between RELEASE-2023 and RELEASE-2024. "
                            "Check data carefully. The recommended workflow is to stack the carbon and water isotope data separately. ")
                
    # set up tables for data and metadata
    # stack months within each site
    sites = np.unique([re.search('\\.([A-Z]+)\\.', key).group(1) for key in verMergDict.keys()])

    # which attributes should be extracted?
    vNames = np.unique([list(verMergDict[key].columns) for key in verMergDict.keys()])
    if metadata:
        if any('rtioMoleDryCo2Vali' in vN for vN in vNames):
            varNames = np.concatenate((sites, ['variables', 'objDesc', 'siteAttributes', 'codeAttributes', 'validationAttributes', 'issueLog', 'scienceReviewFlags']))
            numTabs = 7
        else:
            varNames = np.concatenate((sites, ['variables', 'objDesc', 'siteAttributes', 'codeAttributes', 'issueLog', 'scienceReviewFlags']))
            numTabs = 6
    else:
        varNames = np.concatenate((sites, ['variables', 'objDesc', 'issueLog', 'scienceReviewFlags']))
        numTabs = 4

    # set up final dict
    varMergDict = {}
    for name in varNames:
        varMergDict[str(name)] = []

    for site in tqdm(sites, desc = 'Stacking files by month'):
        mergList = [value for key, value in verMergDict.items() if site in key]
        varMergDict[site] = pd.concat(mergList, ignore_index = True)
        
        # sort by site, hor and ver, and date
        if 'verticalPosition' in varMergDict[site].columns:
            varMergDict[site] = varMergDict[site].sort_values(by=['horizontalPosition', 'verticalPosition', 'timeBgn'], ignore_index=True)
        else:
            varMergDict[site] = varMergDict[site].sort_values(by=['timeBgn'], ignore_index=True)
            
    # attributes, objDesc, SRF table, and issue log
    # attributes are only included if metadata == TRUE
    if metadata:
        siteAttr = []
        codeAttr = []
        for site in tqdm(sites, desc='Getting metadata tables'):
            for file_ in files:
                if re.search(site, file_):
                    sAttr = get_attributes(file_, site, attType='site')
                    siteAttr.append(sAttr)
                    cAttr = get_attributes(file_, site, attType='root')
                    codeAttr.append(cAttr)      
        siteAttributes = pd.DataFrame(siteAttr).map(lambda x: x[0] if isinstance(x, list) and len(x) == 1 else ', '.join(map(str, x)))
        codeAttributes = pd.DataFrame(codeAttr).map(lambda x: x[0] if isinstance(x, list) and len(x) == 1 else ', '.join(map(str, x)))
        varMergDict['siteAttributes'] = siteAttributes
        varMergDict['codeAttributes'] = codeAttributes
        # get CO2 validation attributes if CO2 variables were extracted 
        if 'validationAttributes' in varNames: 
            valAttr = []
            for site in tqdm(sites, desc='Getting validation metadata tables'):
                for okey, idict in tableDict.items():
                    for ikey in idict:
                        if 'rtioMoleDryCo2Vali' and site in ikey:
                            vAttr = get_attributes(okey+'.h5', site, 'val', valName=ikey)
                            valAttr.append(vAttr)
            valAttributes = pd.DataFrame(valAttr).map(lambda x: x[0] if isinstance(x, list) and len(x) == 1 else ', '.join(map(str, x)))
            varMergDict['validationAttributes'] = valAttributes

    # get one objDesc table and add it and variables table to dict
    try:
        with h5py.File(os.path.join(filepath, files[0]), 'r') as h5:
            objDesc = h5['//objDesc'][:]
            objDesc = pd.DataFrame([(x[0].decode('utf-8'), x[1].decode('utf-8')) for x in objDesc], columns = ['Object', 'Description'])
    except Exception:
        # if processing gets this far without failing, don't fail here, just return data without objDesc table
        objDesc = None

    varMergDict['variables'] = variables
    varMergDict['objDesc'] = objDesc

    # get issue log
    if not runLocal:
        try:
            varMergDict['issueLog'] = get_eddy_issue_log(dpid = 'DP4.00200.001')
        except Exception:
            logging.info('Issue log file not accessed - issue log can be found on the data product details pages.')
            varMergDict.pop('issueLog', None)
    else:
        varMergDict.pop('issueLog', None)

    # aggregate the science_review_flags files
    if len(scienceReviewList) > 0:
        outputScienceReview = pd.DataFrame()
        for srFile in scienceReviewList:
            srTemp = pd.read_csv(os.path.join(filepath, srFile))
            outputScienceReview = pd.concat([outputScienceReview, srTemp], ignore_index = True)
        # remove duplicates
        outputScienceReview.drop_duplicates(inplace=True, ignore_index=True)

        # check for non-identical duplicates with the same ID and keep the most recent one
        if len(outputScienceReview['srfID'].unique()) != len(outputScienceReview):
            outputScienceReview['lastUpdateDateTime_s'] = pd.to_datetime(outputScienceReview['lastUpdateDateTime'])
            outputScienceReview = outputScienceReview.sort_values(by=['srfID', 'Time'], ascending = [True, False])
            outputScienceReview.drop_duplicates(subset='srfID', keep='first', inplace=True)
            outputScienceReview.drop(columns=['lastUpdateDateTime_s'], inplace=True)

        varMergDict['scienceReviewFlags'] = outputScienceReview

    else:
        varMergDict.pop('scienceReviewFlags', None)

    # add citations to output
    if not runLocal:
        if citP is not None:
            varMergDict['citation_00200_PROVISIONAL'] = citP
        if citR is not None:
            keyString = 'citation_00200_' + rel[0]
            varMergDict[keyString] = citR

    return varMergDict


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import os
import importlib_resources
import pandas as pd
import logging
from .helper_mods.api_helpers import get_api
from .helper_mods.api_helpers import token_check
from .helper_mods.api_helpers import get_zip_urls
from .helper_mods.api_helpers import get_tab_urls
from .helper_mods.api_helpers import download_urls
from .helper_mods.metadata_helpers import convert_byte_size
from . import __resources__

logging.basicConfig(level=logging.INFO, format="%(message)s")


def query_files(
    lst,
    dpid,
    site="all",
    startdate=None,
    enddate=None,
    package="basic",
    release="current",
    timeindex="all",
    tabl="all",
    include_provisional=False,
    token=None,
):
    """
    Use the query endpoint to get a list of data package files from NEON.

    Parameters
    --------
    lst: Availability info from a call to the products endpoint of the NEON API.
    dpid: Data product identifier in the form DP#.#####.###
    site: Either the string 'all', or one or more 4-letter NEON site codes. Defaults to 'all'.
    startdate: Earliest date of data to download, in the form YYYY-MM
    enddate: Latest date of data to download, in the form YYYY-MM
    package: Download package to access, either basic or expanded
    release: Data release to download. Defaults to the most recent release.
    timeindex: Either the string 'all', or the time index of data to download, in minutes. Only applicable to sensor (IS) data. Defaults to 'all'.
    tabl: Either the string 'all', or the name of a single data table to download. Defaults to 'all'.
    include_provisional: Should Provisional data be returned in the download? Defaults to False.
    token: User specific API token (generated within neon.datascience user accounts). If omitted, download uses the public rate limit.

    Return
    --------
    A list of paths to NEON data package files.

    Created on 10 Jul 2024

    @author: Claire Lunch
    """

    adict = lst["data"]["siteCodes"]
    releasedict = {}
    
    # check for expired token
    if token is not None:
        token = token_check(token)

    # check expanded package status
    if package == "expanded":
        if not lst["data"]["productHasExpanded"]:
            logging.info(
                "No expanded package found for NEON "
                + dpid
                + ". Basic package downloaded instead."
            )
            package = "basic"

    # if sites are not specified, get list of sites with data
    if site == ["all"]:
        siteset = []
        for i in range(0, len(adict)):
            siteset.append(adict[i].get("siteCode"))
    else:
        if isinstance(site, list):
            siteset = site
        else:
            siteset = [site]

    # set up site query
    sitesurllist = ["&siteCode=" + s for s in siteset]
    sitesurl = "".join(sitesurllist)

    # if dates are not specified, get data date range
    if startdate is None or enddate is None:
        dateset = []
        for i in range(0, len(adict)):
            dateset.append(adict[i].get("availableMonths"))
        dateset = sum(dateset, [])
        if startdate is None:
            startdate = min(dateset)
        if enddate is None:
            enddate = max(dateset)

    # set up date query
    dateurl = "&startDateMonth=" + startdate + "&endDateMonth=" + enddate

    # string for true/false include provisional
    if include_provisional is True:
        ipurl = "&includeProvisional=true"
    else:
        ipurl = "&includeProvisional=false"

    if release == "current":
        relurl = ""
    else:
        relurl = "&release=" + release

    # construct full query url and run query
    qurl = (
        "https://data.neonscience.org/api/v0/data/query?productCode="
        + dpid
        + sitesurl
        + dateurl
        + ipurl
        + "&package="
        + package
        + relurl
    )
    qreq = get_api(api_url=qurl, token=token)
    if qreq is None:
        logging.info("No NEON Portal API response for selected query. Check inputs.")
        return None
    qdict = qreq.json()

    # get file list from dictionary response
    reldict = qdict.get("data")
    pdict = reldict.get("releases")
    flurl = []
    for i in range(0, len(pdict)):
        packdict = pdict[i].get("packages")
        rdict = pdict[i].get("release")
        for j in range(0, len(packdict)):
            fdict = packdict[j].get("files")
            for k in range(0, len(fdict)):
                flurl.append(fdict[k].get("url"))
                releasedict[
                    re.sub(
                        pattern="https://storage.googleapis.com/",
                        repl="",
                        string=fdict[k].get("url"),
                    )
                ] = rdict

    # if timeindex or tabl are set, subset the list of files
    if timeindex == "all" and tabl == "all":
        return [flurl, releasedict]
    else:
        if timeindex != "all" and tabl != "all":
            raise ValueError(
                "Only one of timeindex or tabl can be specified, not both."
            )
        else:
            if timeindex != "all":
                tt = re.compile(
                    str(timeindex)
                    + "min|"
                    + str(timeindex)
                    + "_min|science_review_flags|variables|readme|sensor_positions|categoricalCodes"
                )

            if tabl != "all":
                tt = re.compile(
                    "[.]"
                    + tabl
                    + "[.]|variables|readme|sensor_positions|categoricalCodes"
                )

            flurlsub = [f for f in flurl if tt.search(f)]
            releasedictsub = {r: v for r, v in releasedict.items() if tt.search(r)}
            return [flurlsub, releasedictsub]


def zips_by_product(
    dpid,
    site="all",
    startdate=None,
    enddate=None,
    package="basic",
    release="current",
    timeindex="all",
    tabl="all",
    check_size=True,
    include_provisional=False,
    cloud_mode=False,
    progress=True,
    token=None,
    savepath=None,
):
    """
    This function queries the NEON API for data by data product, site(s), and
    month(s), and downloads the corresponding data packages. Use this function
    to download NEON observational (OS) and instrument (IS) data; for remote
    sensing data, use the by_file_aop() and by_tile_aop() functions.

    Parameters
    ------------------
    dpid: str
        Data product identifier in the form DP#.#####.###

    site: str
        Either the string 'all', or one or more 4-letter NEON site codes. Defaults to 'all'.

    startdate: str, optional
        Earliest date of data to download, in the form YYYY-MM

    enddate: str, optional
        Latest date of data to download, in the form YYYY-MM

    package: str, optional
        Download package to access, either basic or expanded. Defaults to 'basic'.

    release: str, optional
        Data release to download. Defaults to the most recent release.

    timeindex: str, optional
        Either 'all', or the time index of data to download, in minutes. Only applicable to sensor (IS) data. Defaults to 'all'.

    tabl: str, optional
        Either the string 'all', or the name of a single data table to download. Only applicable to observational (OS) data. Defaults to 'all'.

    check_size: bool, optional
        True or False, should the user approve the total file size before downloading? Defaults to True. When working in batch mode, or other non-interactive workflow, use check_size=False.

    include_provisional: bool, optional
        Should Provisional data be returned in the download? Defaults to False. See
        https://www.neonscience.org/data-samples/data-management/data-revisions-releases
        for details on the difference between provisional and released data.

    cloud_mode: bool, optional
        Use cloud mode to transfer files cloud-to-cloud? If used, zips_by_product() returns a
        list of files rather than downloading them. Defaults to False; in general this
        option should be used via load_by_product(), in which zips_by_product() is a
        helper function.

    progress: bool, optional
        Should the function display progress bars as it runs? Defaults to True.

    token: str, optional
        User-specific API token from data.neonscience.org user account. See
        https://data.neonscience.org/data-api/rate-limiting/ for details about
        API rate limits and user tokens. If omitted, download uses the public rate limit.

    savepath: str, optional
        File path of location to save data.

    Return
    -----------------
    A folder on the local drive containing the set of data package files meeting the input criteria or,
    in cloud mode, a list of file paths meeting the input criteria.

    Example
    -----------------
    Download water quality data from COMO (Como Creek) in 2018

    >>> zips_by_product(dpid="DP1.20288.001",site="COMO",
                        startdate="2018-01", enddate="2018-12",
                        token=None, savepath="/mypath/Downloads")

    Created on Wed Jan 31 14:36:22 2024

    @author: Claire Lunch
    """

    # error message if dpid is not formatted correctly
    if not re.search(pattern="DP[1-4]{1}.[0-9]{5}.00[0-9]{1}", string=dpid):
        raise ValueError(
            f"{dpid} is not a properly formatted NEON data product ID. The correct format is DP#.#####.00#"
        )

    # error message if package is not basic or expanded
    if not package in ["basic", "expanded"]:
        raise ValueError(
            f"{package} is not a valid NEON download package name. Package must be basic or expanded"
        )

    # error messages for products that can't be downloaded by zips_by_product()
    # AOP products
    if dpid[4:5:1] == 3 and dpid != "DP1.30012.001":
        raise ValueError(
            f"{dpid} is a NEON remote sensing data product. Use the by_file_aop() or by_tile_aop() function."
        )

    # Phenocam products
    if dpid == "DP1.00033.001" or dpid == "DP1.00042.001":
        raise ValueError(
            f"{dpid} is a NEON phenological image product, data are hosted by Phenocam."
        )

    # Aeronet product
    if dpid == "DP1.00043.001":
        raise ValueError(
            f"NEON Spectral sun photometer ({dpid}) data are hosted by Aeronet."
        )

    # DHP expanded package
    if dpid == "DP1.10017.001" and package == "expanded":
        raise ValueError(
            "NEON Digital hemispherical images expanded file packages exceed programmatic download limits. Either download from the data portal, or download the basic package and use the URLs in the data to download the images themselves. Follow instructions in the Data Product User Guide for image file naming."
        )

    # individual SAE products
    if dpid in [
        "DP1.00007.001",
        "DP1.00010.001",
        "DP1.00034.001",
        "DP1.00035.001",
        "DP1.00036.001",
        "DP1.00037.001",
        "DP1.00099.001",
        "DP1.00100.001",
        "DP2.00008.001",
        "DP2.00009.001",
        "DP2.00024.001",
        "DP3.00008.001",
        "DP3.00009.001",
        "DP3.00010.001",
        "DP4.00002.001",
        "DP4.00007.001",
        "DP4.00067.001",
        "DP4.00137.001",
        "DP4.00201.001",
        "DP1.00030.001",
    ]:
        raise ValueError(
            f"{dpid} is only available in the bundled NEON eddy covariance data product. Download DP4.00200.001 to access these data."
        )

    # check for incompatible values of release= and include_provisional=
    if release == "PROVISIONAL" and not include_provisional:
        raise ValueError(
            "NEON Download request is for release=PROVISIONAL. To download PROVISIONAL data, enter input parameter include_provisional=True."
        )
    if re.search(pattern="RELEASE", string=release) is not None and include_provisional:
        logging.info(
            f"NEON Download request is for release={release} but include_provisional=True. Only data in {release} will be downloaded."
        )

    # error message if dates aren't formatted correctly
    # separate logic for each, to easily allow only one to be NA
    if startdate is not None:
        if re.search(pattern="[0-9]{4}-[0-9]{2}", string=startdate) is None:
            raise ValueError(
                "startdate and enddate must be either None or valid dates in the form YYYY-MM"
            )

    if enddate is not None:
        if re.search(pattern="[0-9]{4}-[0-9]{2}", string=enddate) is None:
            raise ValueError(
                "startdate and enddate must be either None or valid dates in the form YYYY-MM"
            )

    # can only specify timeindex xor tabl
    if timeindex != "all" and tabl != "all":
        raise ValueError("Only one of timeindex or tabl can be specified, not both.")
    # consider adding warning message about using tabl=

    # allow for single sites
    if not isinstance(site, list):
        site = [site]

    # redirect for aqu met products and bundles
    shared_aquatic_file = (
        importlib_resources.files(__resources__) / "shared_aquatic.csv"
    )
    shared_aquatic_df = pd.read_csv(shared_aquatic_file, index_col="site")

    if site != ["all"]:
        siter = []
        indx = 0
        for s in site:
            if s in shared_aquatic_df.index:
                if s in ["BLWA"] and release not in ["RELEASE-2021","RELEASE-2022","RELEASE-2023","RELEASE-2024","RELEASE-2025"]:
                    indx = indx + 1
                    if "DELA" not in site:
                        siter.append(["DELA"])
                    if indx == 1:
                        logging.info(
                            f"Some NEON sites in your download request are aquatic sites where {dpid} is collected at a nearby terrestrial site. The sites you requested, and the sites that will be accessed instead, are listed below."
                            )
                    logging.info("Until the fall of 2025, meteorological data for BLWA were collected at DELA. Data collection at DELA ended in late 2025 and the meteorological station was relocated to BLWA. If your download request crosses this time period, data will be downloaded from each site for the time period when they are available.")
                else:
                    ss = shared_aquatic_df.loc[s]
                    if dpid in list(ss["product"]):
                        indx = indx + 1
                        sx = list(ss["towerSite"][ss["product"] == dpid])
                        siter.append(sx)
                        if indx == 1:
                            logging.info(
                                f"Some NEON sites in your download request are aquatic sites where {dpid} is collected at a nearby terrestrial site. The sites you requested, and the sites that will be accessed instead, are listed below."
                                )
                        logging.info(f"{s} -> {''.join(sx)}")
                    else:
                        siter.append([s])
            else:
                siter.append([s])
        siter = sum(siter, [])
    else:
        siter = site

    # redirect for chemistry bundles
    chem_bundles_file = importlib_resources.files(__resources__) / "chem_bundles.csv"
    chem_bundles_df = pd.read_csv(chem_bundles_file)
    if dpid in list(chem_bundles_df["product"]):
        newDPID = list(
            chem_bundles_df["homeProduct"][chem_bundles_df["product"] == dpid]
        )
        if newDPID == ["depends"]:
            raise ValueError(
                "NEON Root chemistry and isotopes have been bundled with the root biomass data. For root chemistry from Megapits, download DP1.10066.001. For root chemistry from periodic sampling, download DP1.10067.001."
            )
        else:
            raise ValueError(
                f"{''.join(dpid)} has been bundled with {''.join(newDPID)} and is not available independently. Please download {''.join(newDPID)}."
            )

    # redirect for veg structure and sediment data product bundles
    other_bundles_file = importlib_resources.files(__resources__) / "other_bundles.csv"
    other_bundles_df = pd.read_csv(other_bundles_file)
    if dpid in list(other_bundles_df["product"]):
        bundle_release = other_bundles_df["lastRelease"][
            other_bundles_df["product"] == dpid
        ].values[0]
        if release > bundle_release:
            newDPID = list(
                other_bundles_df["homeProduct"][other_bundles_df["product"] == dpid]
            )
            raise ValueError(
                f"In all NEON releases after {bundle_release}, {''.join(dpid)} has been bundled with {''.join(newDPID)} and is not available independently. Please download {''.join(newDPID)}."
            )

    # check for expired token
    if token is not None:
        token = token_check(token)

    # end of error and exception handling, start the work
    # query the /products endpoint for the product requested
    if release == "current" or release == "PROVISIONAL":
        prodreq = get_api(
            api_url="https://data.neonscience.org/api/v0/products/" + dpid, token=token
        )
    else:
        prodreq = get_api(
            api_url="https://data.neonscience.org/api/v0/products/"
            + dpid
            + "?release="
            + release,
            token=token,
        )

    if prodreq is None:
        if release == "LATEST":
            logging.info(
                f"No data found for NEON data product {dpid}. LATEST data requested; check that token is valid for LATEST access."
            )
            return
        else:
            if release != "current" and release != "PROVISIONAL":
                rels = get_api(
                    api_url="https://data.neonscience.org/api/v0/releases/", token=token
                )
                if rels is None:
                    raise ConnectionError(
                        f"NEON Data product {dpid} was not found or API was unreachable."
                    )
                relj = rels.json()
                reld = relj["data"]
                rellist = []
                for i in range(0, len(reld)):
                    rellist.append(reld[i]["release"])
                if release not in rellist:
                    raise ValueError(f"Release not found. Valid NEON data releases are {rellist}")
                else:
                    raise ConnectionError(
                        f"NEON Data product {dpid} was not found or API was unreachable."
                    )
            else:
                raise ConnectionError(
                    f"NEON Data product {dpid} was not found or API was unreachable."
                )

    avail = prodreq.json()

    # error message if product or data not found
    # I think this would never be called due to the way get_api() is set up
    try:
        avail["error"]["status"]
        logging.info(f"No NEON data found for data product {dpid}")
        return
    except Exception:
        pass

    # check that token was used
    if "x-ratelimit-limit" in prodreq.headers and token is not None:
        if prodreq.headers.get("x-ratelimit-limit") == 200:
            logging.info("NEON Portal API token was not recognized. Public rate limit applied.")

    # use query endpoint if cloud mode selected
    if cloud_mode:
        fls = query_files(
            lst=avail,
            dpid=dpid,
            site=siter,
            startdate=startdate,
            enddate=enddate,
            package=package,
            release=release,
            timeindex=timeindex,
            tabl=tabl,
            include_provisional=include_provisional,
            token=token,
        )
        return fls

    else:
        # get data urls
        month_urls = []
        for i in range(0, len(avail["data"]["siteCodes"])):
            month_urls.append(avail["data"]["siteCodes"][i]["availableDataUrls"])

        # check for no results
        if len(month_urls) == 0:
            logging.info(f"There are no NEON {dpid} data matching the search criteria.")
            return

        # un-nest list
        month_urls = sum(month_urls, [])

        # subset by site
        if siter != ["all"]:
            site_urls = []
            for si in siter:
                se = re.compile(si)
                month_sub = [s for s in month_urls if se.search(s)]
                site_urls.append(month_sub)
            site_urls = sum(site_urls, [])
        else:
            site_urls = month_urls

        # check for no results
        if len(site_urls) == 0:
            logging.info(f"There are no NEON {dpid} data at the selected sites.")
            return

        # subset by start date
        if startdate is not None:
            ste = re.compile("20[0-9]{2}-[0-9]{2}")
            start_urls = [
                st for st in site_urls if ste.search(st).group(0) >= startdate
            ]
        else:
            start_urls = site_urls

        # check for no results
        if len(start_urls) == 0:
            logging.info(f"There are no NEON {dpid} data at the selected date(s).")
            return

        # subset by end date
        if enddate is not None:
            ete = re.compile("20[0-9]{2}-[0-9]{2}")
            end_urls = [et for et in start_urls if ete.search(et).group(0) <= enddate]
        else:
            end_urls = start_urls

        # check for no results
        if len(end_urls) == 0:
            logging.info(f"There are no NEON {dpid} data at the selected date(s).")
            return

        # if downloading entire site-months, pass to get_zip_urls to query each month for url
        if timeindex == "all" and tabl == "all":
            durls = get_zip_urls(
                url_set=end_urls,
                package=package,
                release=release,
                include_provisional=include_provisional,
                token=token,
                progress=progress,
            )
        else:
            # if downloading by table or averaging interval, pass to get_tab_urls
            durls = get_tab_urls(
                url_set=end_urls,
                package=package,
                release=release,
                include_provisional=include_provisional,
                timeindex=timeindex,
                tabl=tabl,
                token=token,
                progress=progress,
            )

        # check download size
        download_size = convert_byte_size(sum(durls["sz"]))
        if check_size:
            if (
                input(
                    f"Continuing will download {len(durls['z'])} NEON {dpid} files totaling approximately {download_size}. Do you want to proceed? (y/n) "
                )
                != "y"
            ):
                logging.info("Download halted.")
                return None
        else:
            logging.info(
                f"Downloading {len(durls['z'])} NEON {dpid} files totaling approximately {download_size}."
            )

        # set up folder to save to
        if savepath is None:
            savepath = os.getcwd()
        outpath = savepath + "/filesToStack" + dpid[4:9] + "/"

        if not os.path.exists(outpath):
            os.makedirs(outpath)
        else:
            logging.info(
                "Warning: Download folder already exists. Check carefully for duplicate files."
            )

        if timeindex != "all" or tabl != "all":
            for f in durls["flpth"]:
                if not os.path.exists(outpath + f):
                    os.makedirs(outpath + f)

        # download data from each url
        download_urls(url_set=durls, outpath=outpath, token=token, progress=progress)

        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from datetime import datetime
import requests


def get_citation(dpid, release):
    """
    Use the DOI Foundation API to get BibTex-formatted citations for NEON data,
    or use a template to generate a BibTex citation for provisional data.
    Helper function to download and stacking functions.

    Parameters
    ----------
    dpid: str
        The data product ID of the data to be cited

    release: str
        The data release to be cited. Can be provisional.

    Return
    ------
    A character string containing the citation in BibTex format.

    Example
    -------
    Get the citation for Breeding landbird point counts (DP1.10003.001),
    RELEASE-2023

    >>> cit = get_citation(dpid="DP1.10003.001", release="RELEASE-2023")

    Created on Fri Aug 25 10:26:23 2023

    @author: Claire Lunch
    """

    if release == "PROVISIONAL":
        # construct citation from template
        citI = "@misc{DPID/provisional,\n  doi = {},\n  url = {https://data.neonscience.org/data-products/DPID},\n  author = {{National Ecological Observatory Network (NEON)}},\n  language = {en},\n  title = {NAME (DPID)},\n  publisher = {National Ecological Observatory Network (NEON)},\n  year = {YEAR}\n}"
        citDP = citI.replace("DPID", dpid)
        citY = citDP.replace("YEAR", str(datetime.now().year))

        nm_req = requests.get("https://data.neonscience.org/api/v0/products/" + dpid)
        nm_str = nm_req.json()
        nm = nm_str["data"]["productName"]

        cit = citY.replace("NAME", nm)
        return cit

    else:
        # get DOI from NEON API, then citation from DOI API
        pr_req = requests.get("https://data.neonscience.org/api/v0/products/" + dpid)
        pr_str = pr_req.json()
        rels = pr_str["data"]["releases"]
        relinfo = next((i for i in rels if i["release"] == release), None)

        if relinfo is None:
            print("There are no data with dpid=" + dpid + " and release=" + release)
            return relinfo

        else:
            doi = relinfo["productDoi"]["url"]
            doi_req = requests.get(doi, headers={"accept": "application/x-bibtex"})
            return doi_req.text



# -*- coding: utf-8 -*-
"""
Created on Thu Apr 11 15:17:04 2024

@author: bhass

Unit tests for get_issue_log

These will test static part of issue logs, since change and issue logs are subject to change.

Notes:
- These tests require an internet connection to run, as they are actually making API calls

"""

import unittest
from src.neonutilities.get_issue_log import get_issue_log


class TestGetIssueLog(unittest.TestCase):
    def test_get_issue_log_chm(self):
        # Call the function for CHM
        df = get_issue_log('DP3.30015.001')
        # Check the column names
        self.assertEqual(list(df.columns), ['id', 'parentIssueID', 'issueDate',
                                            'resolvedDate', 'dateRangeStart',
                                            'dateRangeEnd', 'locationAffected',
                                            'issue', 'resolution'])

        self.assertEqual(
            df.loc[0, 'issue'][0:57], 'Original algorithm did not take advantage of improvements')
        self.assertEqual(df.loc[0, 'issueDate'], '2022-01-19T00:00:00Z')

    def test_get_issue_log_eddy(self):
        # Call the function for bundled eddy covariance
        df = get_issue_log('DP4.00200.001')

        # Check the column names
        self.assertEqual(list(df.columns), ['dpid', 'id', 'parentIssueID', 'issueDate',
                                            'resolvedDate', 'dateRangeStart',
                                            'dateRangeEnd', 'locationAffected',
                                            'issue', 'resolution'])

        # # Check a few rows
        # # Replace with the expected value
        # self.assertEqual(df.loc[0, 'col1'], 'expected value')
        # # Replace with the expected value
        # self.assertEqual(df.loc[10, 'col2'], 'expected value')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import requests
import os
import platform
import logging
import importlib.metadata
import pandas as pd
from tqdm import tqdm
from urllib.parse import urlparse
from .unzip_and_stack import unzip_zipfile
from .helper_mods.metadata_helpers import convert_byte_size

# Set global user agent
vers = importlib.metadata.version("neonutilities")
plat = platform.python_version()
osplat = platform.platform()
usera = f"neonutilities/{vers} Python/{plat} {osplat}"

def files_by_uri(
    filepath,
    savepath=None,
    check_size=True,
    unzip=True,
    save_zipped_files=False,
    progress=True,
):
    """

    Get files from NEON GCS Bucket using URLs in stacked data

    Parameters
    --------
    filepath: The location of the NEON data containing URIs. Can be either a local directory containing NEON tabular data or a list object containing tabular data.
    savepath: The location to save the output files from the GCS bucket, optional. Defaults to creating a "GCS_zipFiles" folder in the filepath directory.
    check_size: should the user be told the total file size before downloading, defaults to True? (true/false)
    unzip: indicates if the downloaded files from GCS buckets should be unzipped into the same directory, defaults to True. Supports .zip and .tar.gz files currently. (true/false)
    save_zipped_files: Should the unzipped monthly data folders be retained, defaults to False? Supports .zip and .tar.gz files currently. (true/false)
    progress: Should a progress bar be displayed?

    Return
    --------
    A folder in the working directory (or in savepath, if specified), containing all files meeting query criteria.

    Example
    --------
    ZN NOTE: Insert example when function is coded

    >>> example

    Created on Fri Aug 9 2024

    @author: Zachary Nickerson
    """

    # check that filepath points to either a directory or a Python list object
    if not isinstance(filepath, dict):
        if not os.path.exists(filepath):
            raise Exception(
                "Input filepath is not a list object in the environment nor an existing file directory."
            )

    # if filepath is a dictionary, make a savepath
    if isinstance(filepath, dict):
        tabList = filepath
        if savepath is not None:
            if not os.path.exists(savepath):
                try:
                    os.makedirs(os.path.join(savepath, "GCS_files"))
                except OSError:
                    print(
                        f"Could not create savepath directory. NEON files will be saved to {os.getcwd()}/GCS_files"
                    )
                    savepath = f"{os.getcwd()}"
    else:
        # if filepath is a directory, read in contents
        files = [
            os.path.join(filepath, f)
            for f in os.listdir(filepath)
            if os.path.isfile(os.path.join(filepath, f))
        ]
        files = [file for file in files if file.endswith(".csv")]
        tabList = {}
        for j, file in enumerate(files):
            try:
                tabList[file] = pd.read_csv(file)
            except Exception:
                print(f"File {file} could not be read.")
                tabList[f"error{j}"] = None
                continue
        tabList = {k: v for k, v in tabList.items() if not k.startswith("error")}

    # Check for the variables file in the filepath
    varList = [k for k in tabList.keys() if "variables" in k]
    if len(varList) == 0:
        raise Exception("NEON Variables file was not found in specified filepath.")
    if len(varList) > 1:
        raise Exception("More than one NEON variables file found in filepath.")
    varFile = tabList[varList[0]]

    URLs = varFile[varFile["dataType"] == "uri"]

    # All of the tables in the package with URLs to download
    allTables = URLs["table"].unique()

    # Loop through tables and fields to compile a list of URLs to download
    URLsToDownload = []

    # Remove allTables values that aren't in tabList
    allTables = {key for key in tabList.keys() for substr in allTables if substr in key}

    if len(allTables) < 1:
        raise Exception("No NEON tables with URIs available in download package contents.")

    # Loop through tables
    for table in allTables:
        tableData = tabList[table]
        # Find URLs per table that are in URLs.fieldName
        URLsPerTable = [url for url in tableData if url in URLs["fieldName"].values]
        # Append the URLs from the fields found
        URLsToDownload.extend([tableData[url] for url in URLsPerTable])

    # Remove duplicates from the list of URLs
    uniqueURLs = set()
    for lst in URLsToDownload:
        uniqueURLs.update(lst)
    URLsToDownload = uniqueURLs

    # Remove None values
    URLsToDownload = [url for url in URLsToDownload if url is not None]
    URLsToDownload = [url for url in URLsToDownload if str(url) != "nan"]
    URLsToDownload = [url for url in URLsToDownload if str(url) != ""]

    if len(URLsToDownload) == 0:
        raise Exception("There are no NEON URLs other than 'None' for the stacked data.")

    # Create savepath only if it does not already exist
    if savepath is not None:
        savepath = os.path.join(savepath, "GCS_files")
    else:
        print(
            f"Could not create savepath directory. NEON files will be saved to {os.getcwd()}/GCS_files"
        )
        savepath = f"{os.getcwd()}/GCS_files"
    if not os.path.exists(os.path.join(savepath)):
        os.makedirs(savepath)        

    # Check the existence and size of each file from URL
    if check_size:
        logging.info(f"Checking size of downloading {len(URLsToDownload)} files by URI")
        sz = []
        for urlfile in tqdm(URLsToDownload, disable=not progress):
            response = requests.head(
                urlfile, headers={"User-Agent": usera}, allow_redirects=True
            )

            # Return nothing if response failed
            if response.status_code != 200:
                logging.info(
                    "Connection error for a subset of urls. Check outputs for missing data."
                )
                # return None

            # Compile file sizes
            flszi = int(response.headers["Content-Length"])
            sz.append(flszi)

        # Check download size
        download_size = convert_byte_size(sum(sz))
        if (
            input(
                f"Continuing will download {len(URLsToDownload)} NEON files totaling approximately {download_size}. Do you want to proceed? (y/n) "
            )
            != "y"
        ):
            logging.info("Download halted.")
            return None
        else:
            logging.info(
                f"Downloading {len(URLsToDownload)} NEON files totaling approximately {download_size}."
            )

    # Download URLs
    for j in tqdm(URLsToDownload, disable=not progress):
        parsed_url = urlparse(j)
        filename = os.path.basename(parsed_url.path)
        file_path = os.path.join(savepath, filename)
        # Download the file
        response = requests.get(j, headers={"User-Agent": usera})
        with open(file_path, "wb") as out_file:
            out_file.write(response.content)
        # If the file type is zip and unzip is True, unzip
        if unzip is True:
            if file_path.endswith(".zip"):
                unzip_zipfile(zippath=file_path)
                # Remove zip files after being unzipped
                if save_zipped_files is False:
                    os.remove(file_path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
import pyarrow as pa
from pyarrow import dataset
import logging

logging.basicConfig(level=logging.INFO, format="%(message)s")


def get_variables(v):
    """

    Get correct data types for each field in a data table

    Parameters
    --------
    v: A pandas table containing variable definitions

    Return
    --------
    A pyarrow schema for data types based on the variables file

    Created on Wed Apr 17 2024

    @author: Zachary Nickerson
    """

    # function assumes variables are loaded as a pandas data frame.

    # create pyarrow schema by translating NEON data types to pyarrow types
    for i in range(0, len(v)):
        nm = v.fieldName[i]
        typ = pa.string()
        if v.dataType[i] == "real":
            typ = pa.float64()
        if v.dataType[i] in ["integer", "unsigned integer", "signed integer"]:
            typ = pa.int64()
        if v.dataType[i] in ["string", "uri"]:
            typ = pa.string()
        if v.dataType[i] == "dateTime":
            if v.pubFormat[i] in [
                "yyyy-MM-dd'T'HH:mm:ss'Z'(floor)",
                "yyyy-MM-dd'T'HH:mm:ss'Z'",
                "yyyy-MM-dd'T'HH:mm:ss'Z'(round)",
                "yyyy-MM-dd'T'HH:mm'Z'(floor)",
                "yyyy-MM-dd'T'HH:mm'Z'",
                "yyyy-MM-dd'T'HH:mm'Z'(round)",
                "yyyy-MM-dd'T'HH'Z'(floor)",
                "yyyy-MM-dd'T'HH'Z'",
                "yyyy-MM-dd'T'HH'Z'(round)"
            ]:
                typ = pa.timestamp("s", tz="UTC")
            else:
                if v.pubFormat[i] in ["yyyy-MM-dd'T'HH:mm:ss.SSS'Z'"]:
                    typ = pa.timestamp("ms", tz="UTC")
                else:
                    if v.pubFormat[i] in ["yyyy-MM-dd(floor)", "yyyy-MM-dd"]:
                        typ = pa.date64()
                    else:
                        if v.pubFormat[i] in ["yyyy(floor)", "yyyy(round)"]:
                            typ = pa.int64()
                        else:
                            typ = pa.string()
        if i == 0:
            vschema = pa.schema([(nm, typ)])
        else:
            nfield = pa.field(nm, typ)
            vschema = vschema.append(nfield)

    return vschema


def read_table_neon(data_file, var_file):
    """

    Read a NEON data table with correct data types for each variable.

    Parameters
    -------------------
    data_file: str
        Filepath to a data table to load.

    var_file: str
        Filepath to a variables file matching the data table.

    Return
    -------------------
    A data frame of a NEON data table, with column classes assigned by data type.

    Example
    -------------------
    >>> dattab = read_table_neon(data_file="path to data table file",
                                 var_file="path to variables file matching data table")

    Created on Thu Feb 1 2024

    @author: Zachary Nickerson
    """

    # Read in variables file and check type
    if isinstance(var_file, str):
        try:
            v = pd.read_csv(var_file)
        except Exception:
            logging.info(
                "Table read failed because var_file must be either a NEON variables table or a file path to a NEON variables table."
            )
            return

    # Check this is a valid variables file
    if any(x in ["category", "system", "stat"] for x in list(v.columns)):
        print(
            "var_file appears to match DP4.00200.001. Data wrangling for surface-atmosphere exchange data is currently only available in the R package version of neonUtilities."
        )
        return
    else:
        if not any(x in ["table", "fieldName", "dataType"] for x in list(v.columns)):
            logging.info(
                "var_file is not a variables file, or is missing critical values."
            )
            return

    # get field names from the data table without loading table
    tabcols = pd.read_csv(data_file, index_col=0, nrows=0).columns.tolist()

    # subset variables file to the relevant fields
    vtab = v[v["fieldName"].isin(tabcols)]
    vtabu = vtab.drop_duplicates(subset="fieldName", ignore_index=True)

    # generate schema
    tableschema = get_variables(vtabu)

    # Check that most fields have a corrsponding value in variables
    m = len(tabcols) - len(tableschema)
    if m == len(tabcols):
        logging.info(
            "Cannot correct data types because the variables file does not match the data file."
        )
        return
    if m > 4:
        logging.info(
            f"{m} fieldNames are present in data file but not in variables file. Data load may be affected; if possible, unknown fields are read as character strings."
        )

    # read data and append file names
    dat = dataset.dataset(source=data_file, format="csv", schema=tableschema)
    dattab = dat.to_table()
    pdat = dattab.to_pandas()

    return pdat


def date_convert(dates):
    """

    Convert a string to a date type, or, if possible date formats fail, keep as a string.

    Parameters
    --------
    dates: A date or datetime in string format

    Return
    --------
    The date in date format, or, if conversion fails, the original string.

    Example
    --------
    >>> d = date_convert("2023-08-01")

    Created on 2 May 2024

    @author: Claire Lunch
    """

    try:
        dout = pd.to_datetime(dates, format="ISO8601", utc=True)
    except Exception:
        try:
            dout = pd.to_datetime(dates, format="%Y-%m-%dT%H:%M:%S", utc=True)
        except Exception:
            try:
                dout = pd.to_datetime(dates, format="%Y-%m-%dT%H:%M", utc=True)
            except Exception:
                try:
                    dout = pd.to_datetime(dates, format="%Y-%m-%dT%H", utc=True)
                except Exception:
                    try:
                        dout = pd.to_datetime(dates, format="%Y-%m-%d", utc=True)
                    except Exception:
                        dout = dates
    return dout


def get_variables_pandas(v):
    """

    Get correct data types for each field in a data table

    Parameters
    --------
    v: A pandas table containing variable definitions

    Return
    --------
    A dictionary of field names and pandas data types based on the variables file

    Created on Oct 31 2024

    @author: Claire Lunch
    """

    dtdict = {}
    vf = v["fieldName"]
    for i in range(0, len(vf)):
        nm = vf[i]
        if v["dataType"][i] == "real":
            typ = "Float64"
        if v["dataType"][i] in ["integer", "unsigned integer", "signed integer"]:
            typ = "Int64"
        if v["dataType"][i] in ["string", "uri"]:
            typ = "string"
        if v["dataType"][i] == "dateTime":
            typ = "datetime64[ns, UTC]"
        dtdict[nm] = typ

    return dtdict


def cast_table_neon(data_table, var_table):
    """

    Cast a NEON data table to the correct data types for each variable, if possible.

    Parameters
    --------
    data_table: NEON data table as a pandas table.
    var_table: NEON variables file as a pandas table.

    Return
    --------
    A data frame of a NEON data table, with column classes assigned by data type.

    Example
    --------
    >>> dattab = cast_table_neon(data_table=brd_perpoint,
                                 var_table=variables_10003)

    Created on Oct 30 2024

    @author: Claire Lunch
    """

    # Check inputs formatting
    if not isinstance(data_table, pd.DataFrame):
        logging.info("Data table input is not a pandas data frame.")
        return None

    if not isinstance(var_table, pd.DataFrame):
        logging.info("Variables table input is not a pandas data frame.")
        return None

    # Check this is a valid variables file
    if any(x in ["category", "system", "stat"] for x in list(var_table.columns)):
        logging.info(
            "var_table appears to match DP4.00200.001. Data wrangling for surface-atmosphere exchange data is currently only available in the R package version of neonUtilities."
        )
        return None
    else:
        if not any(
            x in ["table", "fieldName", "dataType"] for x in list(var_table.columns)
        ):
            logging.info(
                "var_table is not a variables file, or is missing critical values."
            )
            return None

    # get data types
    vdt = get_variables_pandas(var_table)

    # get field names from the data table
    tabcols = list(data_table.columns)
    cast_table = data_table

    # iterate over columns and try to cast each
    for i in tabcols:
        if i not in vdt.keys():
            continue
        else:
            if vdt[i] in ["Float64", "Int64"]:
                try:
                    dtemp = cast_table[i].replace(r"^\s*$", np.nan, regex=True)
                    cast_table[i] = dtemp.astype(vdt[i])
                except Exception:
                    logging.info(
                        f"Field {i} could not be cast to type {vdt[i]}. Data read as string type."
                    )
                    cast_table[i] = data_table[i]
                    continue
            if vdt[i] == "datetime64[ns, UTC]" and not i == "publicationDate":
                try:
                    cast_table[i] = date_convert(data_table[i])
                except Exception:
                    logging.info(
                        f"Field {i} could not be cast to type {vdt[i]}. Data read as string type."
                    )
                    cast_table[i] = data_table[i]
                    continue

    return cast_table



from .citation import get_citation
from .aop_download import (
    by_file_aop,
    by_tile_aop,
    list_available_dates,
    get_aop_tile_extents,
)
from .tabular_download import zips_by_product
from .get_issue_log import get_issue_log
from .read_table_neon import read_table_neon
from .helper_mods.api_helpers import token_date
from .unzip_and_stack import (
    stack_by_table,
    load_by_product,
    dataset_query,
)
from .stack_eddy import stack_eddy



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 9 01:34:00 2025

@author: bhass

Unit tests for metadata_helpers.py
"""

import unittest
from src.neonutilities.helper_mods.metadata_helpers import convert_byte_size


class TestConvertByteSize(unittest.TestCase):
    def test_convert_byte_size(self):
        """
        Test the convert_byte_size function with various byte sizes.
        """
        # Test bytes
        self.assertEqual(convert_byte_size(500), "500.0 B")

        # Test kilobytes
        self.assertEqual(convert_byte_size(1024), "1.0 KB")
        self.assertEqual(convert_byte_size(1536), "1.5 KB")

        # Test megabytes
        self.assertEqual(convert_byte_size(1048576), "1.0 MB")
        self.assertEqual(convert_byte_size(1572864), "1.5 MB")

        # Test gigabytes
        self.assertEqual(convert_byte_size(1073741824), "1.0 GB")
        self.assertEqual(convert_byte_size(1610612736), "1.5 GB")

        # Test terabytes
        self.assertEqual(convert_byte_size(1099511627776), "1.0 TB")
        self.assertEqual(convert_byte_size(1649267441664), "1.5 TB")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Created on 14 Apr 2025

@author: Zachary Nickerson

Unit tests for files_by_uri()

"""

# import required packages
from src.neonutilities.files_by_uri import files_by_uri
import os
import shutil

def test_files_by_uri_NEF():
    """
    Test that the function works for NEF files available from DP1.10017.001 (tabular data saved in testdata)
    """
    testdir = "./testdata/NEON_uri_testdata/10017_DELA_202306_RELEASE2025"
    files_by_uri(testdir,
                 savepath=testdir,
                 check_size=False,
                 unzip=True,
                 save_zipped_files=False,
                 progress=True)
    # There should be 36 .NEF files in the directory
    files = os.listdir(os.path.join(testdir,"GCS_files"))
    nef_files = [file for file in files if file.lower().endswith('.nef')]
    assert len(nef_files) == 36
    # Remove the .NEF files saved in the testdir
    shutil.rmtree(os.path.join(testdir,"GCS_files"))

def test_files_by_uri_ZIP():
    """
    Test that the function works for ZIP files available from DP4.00131.001 (tabular data saved in testdata)
    """
    testdir = "./testdata/NEON_uri_testdata/00131_allSites_2022_RELEASE2025"
    files_by_uri(testdir,
                 savepath=testdir,
                 check_size=False,
                 unzip=False,
                 save_zipped_files=False,
                 progress=True)
    # There should be 21 .ZIP files in the directory
    files = os.listdir(os.path.join(testdir,"GCS_files"))
    zip_files = [file for file in files if file.lower().endswith('.zip')]
    assert len(zip_files) == 21
    # Remove the .NEF files saved in the testdir
    shutil.rmtree(os.path.join(testdir,"GCS_files"))    
    
def test_files_by_uri_ZIP_unzip():
    """
    Test that the function works for ZIP files available from DP4.00131.001, and unzip the files (tabular data saved in testdata)
    """
    testdir = "./testdata/NEON_uri_testdata/00131_allSites_2022_RELEASE2025"
    files_by_uri(testdir,
                 savepath=testdir,
                 check_size=False,
                 unzip=True,
                 save_zipped_files=False,
                 progress=True)
    # There should be 294 files that are not .CSV in the directory
    files = os.listdir(os.path.join(testdir,"GCS_files"))
    geo_files = [file for file in files if not file.lower().endswith('.csv')]
    assert len(geo_files) == 294
    # Remove the .NEF files saved in the testdir
    shutil.rmtree(os.path.join(testdir,"GCS_files"))

# Potentially add a test for the microbial file types